/*
 * STM32F4 Wifi flight controller
 * Copyright (C) 2012-2014 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

// Handle protocol between copter & the ground


#include "settings.h"
#include "arm/config.h"
#include "pic/config.h"
#include "stm32f4xx_tim.h"
#include "linux.h"
#include "wifi.h"
#include "rtl_usb.h"

#include "autopilot.h"
#include "bluetooth.h"
#include "camera.h"
#include "copter.h"
#include "copter_io.h"


#ifdef USE_FLOW
#include "flow.h"
#endif

#include "imu.h"
#if defined(PIC_USE_MRF49XA) || defined(PIC_USE_MRF49XA_GROUND)
#include "mrf49xa.h"
#endif

#include "nav.h"
#include "pwm.h"
#include "sonar_air.h"
#include "spi.h"
#include "net.h"
#include "uart.h"
#include "usb.h"
#include "gyro.h"

// Maximum picture fragments per beacon
#define PICTURE_FRAGMENTS 16
#define LEFT_PICTURE_FRAGMENTS 2
// Number of times to resend current picture fragment
#define TOTAL_RESENDS 1
// Minimum time between beacons required by SPI
#define LEFT_BEACON_DELAY 20000

// web interface
#define TEXTLEN 1024

copter_io_t copter_io;



#ifdef USE_WIFI

unsigned char string[TEXTLEN];
const unsigned char *string_end;
int new_channel;
unsigned char new_ssid[HOSTAPD_MAX_SSID_LEN + 1];


void forward_beacon();



// Must be last packet sent after a beacon, to avoid having the ground transmit
// something during our later packet.
void forward_beacon()
{
	unsigned char *buf = net.in_buffer + PACKET_OFFSET;
	int len = net.bytes_received - PACKET_OFFSET;

	memcpy(net.out_buffer, buf, len);

	struct ieee80211_hdr *hdr = (struct ieee80211_hdr*)net.out_buffer;
	
// ground
	memcpy(hdr->addr1, net.ground_mac, ETH_ALEN);
	memcpy(hdr->addr2, get_mac(), ETH_ALEN);
// laptop
	memcpy(hdr->addr3, net.src_mac, ETH_ALEN);


/*
 * TRACE2
 * print_number(len);
 * print_mac(hdr->addr1);
 * print_mac(hdr->addr2);
 * print_mac(hdr->addr3);
 */


	net_start_transmit(net.out_buffer,
 		len,
 		RTL_TXQ_MGT,
		0);


}






typedef struct
{
	unsigned char *path;
	unsigned char *text;
} uri_table_t;

const unsigned char content_ok[] = 
"HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: ";

const unsigned char content_fail[] = 
"HTTP/1.1 404 Not Found\r\nContent-Type: text/html\r\nContent-Length: ";

void send_text(const unsigned char *header, 
	const unsigned char *text)
{
	unsigned char *ptr4 = net.out_buffer;
	unsigned char *out_end = net.out_buffer + ARM_OUT_BUFFER_SIZE;
	int text_len = strlen(text);
	int header_len = strlen(header);
	memcpy(ptr4, header, header_len);
	ptr4 += header_len;

	ptr4 += sprint_number(ptr4, text_len, out_end - ptr4);
	*ptr4++ = '\r';
	*ptr4++ = '\n';
	*ptr4++ = '\r';
	*ptr4++ = '\n';

	memcpy(ptr4, text, text_len);
	ptr4 += text_len;

	int payload_size = ptr4 - net.out_buffer;
	*ptr4++ = 0;
	*ptr4++ = 0;
	send_tcp(payload_size,
		net.sequence_number,
		net.ack_number,
		0x18,
		net_start_receive);
}

unsigned char* print_header(unsigned char *ptr)
{
	ptr += strncpy(ptr, "<H1>MARCY 2 CONFIGURATION</H1><P>\r\n\r\n", string_end - ptr);
	return ptr;
}

static unsigned char* print_config(unsigned char *ptr)
{
	ptr += strncpy(ptr, "CHANNEL: <b>", string_end - ptr);
	ptr += sprint_number(ptr, new_channel, string_end - ptr);


	ptr += strncpy(ptr, 
		"</b><P>\r\n\r\nSSID: <b>", string_end - ptr);
	ptr += strncpy(ptr, new_ssid, string_end - ptr);
	ptr += strncpy(ptr, "</b><p>\r\n\r\n", string_end - ptr);
	return ptr;
}

// root web page
void send_root()
{
	unsigned char *ptr = string;
	ptr = print_header(ptr);
	ptr += strncpy(ptr, 
		"<form name=\"config\" action=\"apply.html\" method=\"get\">\r\n", string_end - ptr);
	ptr += strncpy(ptr, 
		"CHANNEL: <input name=\"channel\" id=\"channel\" value=\"", string_end - ptr);
	ptr += sprint_number(ptr, new_channel, string_end - ptr);
	ptr += strncpy(ptr, "\" size=\"5\" maxlength=\"2\" type=\"text\"><p>\r\n\r\n", string_end - ptr);

	ptr += strncpy(ptr, 
		"SSID: <input name=\"ssid\" id=\"ssid\" value=\"", string_end - ptr);
	ptr += strncpy(ptr, new_ssid, string_end - ptr);
	ptr += strncpy(ptr, "\" size=\"40\" maxlength=\"32\" type=\"text\"><p>\r\n\r\n", string_end - ptr);

	ptr += strncpy(ptr, "<input type=\"submit\" value=\"Apply\"><P>\r\n\r\n", string_end - ptr);
	
	ptr += strncpy(ptr, "</form>\r\n", string_end - ptr);




	ptr += strncpy(ptr, "<hr><p><H1>FACTOIDS</H1><P>\r\n\r\n<b>CURRENTLY ASSOCIATED:</b><p>\r\n", string_end - ptr);
	int i, j;
	for(i = 0; i < net.total_stations; i++)
	{
		unsigned char *mac = net.station_macs[i];
		for(j = 0; j < ETH_ALEN && ptr < string_end; j++)
		{
			if(j > 0)
			{
				if(ptr < string_end) *ptr++ = ':';
			}
			
			if(ptr < string_end) *ptr++ = hex_table[(mac[j] >> 4) & 0xf];
			if(ptr < string_end) *ptr++ = hex_table[mac[j] & 0xf];
		}
		ptr += strncpy(ptr, "<p>\r\n", string_end - ptr);
	}

	ptr += strncpy(ptr, "<b>RX BYTES: </b>", string_end - ptr);
	ptr += sprint_number(ptr, net.total_rx, string_end - ptr);

	ptr += strncpy(ptr, " <b>TX BYTES: </b>", string_end - ptr);
	ptr += sprint_number(ptr, net.total_tx, string_end - ptr);
	ptr += strncpy(ptr, "<p>\r\n", string_end - ptr);





	if(ptr - string > TEXTLEN)
	{
		TRACE
		print_text("text overflow");
	}

	send_text(content_ok, string);
	
}

void send_apply(const unsigned char *uri,
	const unsigned char *uri_end)
{
	unsigned char *ptr = string;


// Read new settings from URI
	new_channel = copter.channel;
	strcpy(new_ssid, copter.ssid);

	const unsigned char *ptr1 = uri;
	while(ptr1 < uri_end - 4)
	{
		if(ptr1[0] == 'n' &&
			ptr1[1] == 'e' &&
			ptr1[2] == 'l' &&
			ptr1[3] == '=')
		{
			new_channel = 0;
			ptr1 += 4;
			if(ptr1 < uri_end && *ptr1 >= '0' && *ptr1  <= '9')
			{
				new_channel = *ptr1 - '0';
				ptr1++;
			}
			
			if(ptr1 < uri_end && *ptr1 >= '0' && *ptr1  <= '9')
			{
				new_channel *= 10;
				new_channel += *ptr1 - '0';
				ptr1++;
			}

			break;
		}

		ptr1++;
	}

	while(ptr1 < uri_end - 4)
	{
		if(ptr1[0] == 's' &&
			ptr1[1] == 'i' &&
			ptr1[2] == 'd' &&
			ptr1[3] == '=')
		{
			ptr1 += 4;
			unsigned char *ptr2 = new_ssid;
			while(ptr1 < uri_end && 
				*ptr1 != ' ' && 
				ptr2 - new_ssid < HOSTAPD_MAX_SSID_LEN)
			{
				*ptr2++ = *ptr1++;
			}
			*ptr2++ = 0;
			break;
		}

		ptr1++;
	}

	CLAMP(new_channel, 1, 12);

	ptr = print_header(ptr);
	ptr += strncpy(ptr, "Confirm your new settings before saving.<P>\r\n\r\n", string_end - ptr);


	ptr = print_config(ptr);


	ptr += strncpy(ptr, 
		"<form name=\"confirm\" action=\"save.html\" method=\"get\">\r\n", string_end - ptr);
	ptr += strncpy(ptr, "<input type=\"submit\" value=\"Apply\"><P>\r\n\r\n", string_end - ptr);
	ptr += strncpy(ptr, "</form>\r\n", string_end - ptr);

	ptr += strncpy(ptr, 
		"<form name=\"confirm\" action=\"index.html\" method=\"get\">\r\n", string_end - ptr);
	ptr += strncpy(ptr, "<input type=\"submit\" value=\"Cancel\"><P>\r\n\r\n", string_end - ptr);
	ptr += strncpy(ptr, "</form>\r\n", string_end - ptr);


	send_text(content_ok, string);
	
}

void send_saved()
{
	unsigned char *ptr = string;

	if(ENGINE_ON)
	{
		ptr = print_header(ptr);
		ptr += strncpy(ptr, "CAN'T CHANGE SETTINGS WHEN MOTOR IS ON.<P>\r\n\r\n", string_end - ptr);
	}
	else
	{
//		net.channel = new_channel;
//		strcpy(net.ssid, new_ssid);
//		net.ssid_len = strlen(new_ssid);

// Write to flash but don't change network, so the response page loads.
// Have to change network to write the settings in 1 function.
		copter.channel = new_channel;
		strcpy(copter.ssid, new_ssid);
		save_settings();
		ptr = string;

// Disable the motor
		net.settings_changed = 1;

		ptr = print_header(ptr);
		ptr += strncpy(ptr, "New settings saved.  Power cycle the aircraft for them to take effect.<P>\r\n\r\n", string_end - ptr);
		ptr += strncpy(ptr, "MOTOR DISABLED UNTIL REBOOT.<P>\r\n\r\n", string_end - ptr);


	}

	ptr = print_config(ptr);

	send_text(content_ok, string);
}

int test_uri(const unsigned char *test_string, 
	const unsigned char *uri,
	const unsigned char *uri_end)
{
	const unsigned char *ptr1 = test_string;
	const unsigned char *ptr3 = uri;
	while(ptr3 < uri_end &&
		(*ptr3 != ' ' && *ptr3 != '?') && 
		*ptr1 != 0 &&
		*ptr1 == *ptr3)
	{
		ptr1++;
		ptr3++;
	}

	if(ptr3 < uri_end &&
		(*ptr3 == ' ' || *ptr3 == '?') &&
		*ptr1 == 0)
	{
		return 1;
	}
	
	return 0;
}

void handle_http(unsigned char *data, int size)
{
// request method
	if(!memcmp(data, "GET", 3))
	{
// URI
		int i;
		unsigned char *data_end = data + size;
		unsigned char *uri_ptr = data + 3;
		
		while(*uri_ptr == ' ' && uri_ptr < data_end) uri_ptr++;

		if(uri_ptr >= data_end)
		{
			net_start_receive();
			return;
		}

		if(test_uri("/", uri_ptr, data_end) ||
			test_uri("/index.html", uri_ptr, data_end))
		{
			send_root();
			return;
		}

		if(test_uri("/apply.html", uri_ptr, data_end))
		{
			send_apply(uri_ptr, data_end);
			return;
		}

		if(test_uri("/save.html", uri_ptr, data_end))
		{
			send_saved();
			return;
		}

		int got_it = 0;
// Compare static URI contents
//		for(i = 0; i < uri_table_size; i++)
//		{
//			if(test_uri(uri_table[i].path, uri_ptr, data_end))
//			{
// Send text
//				send_text(content_ok, uri_table[i].text);
//				return;
//			}
//		}
		
		if(!got_it)
		{
			unsigned char *ptr = string;
			ptr += strncpy(ptr, 
				"<title>404 Not Found</title>\r\n<h1>Not Found</h1>\r\nThe requested URL ",
				string_end - ptr);

			unsigned char *ptr3 = uri_ptr;
			while(ptr3 < data_end &&
				(*ptr3 != ' ' && *ptr3 != '?') &&
				ptr < string_end - 1)
			{
				*ptr++ = *ptr3++;
			}

			ptr += strncpy(ptr, 
				" was not found on this aircraft.<P>\r\n<hr><P>\r\n",
				string_end - ptr);
			if(ptr < string_end) *ptr++ = 0;

			send_text(content_fail, string);
			return;
		}
	}

	net_start_receive();
}

#endif // USE_WIFI
















#if !defined(PACKET_FORWARD) || defined(COPTER_MODE)





#ifdef DEBUG_SONAR2

void send_audio_fragment()
{
//TRACE2
//print_number(sonar.buffer_used);
	if(sonar.write_byte < DEBUG2_PACKET)
	{
		net_start_receive();
		return;
	}

	int payload_size = DEBUG2_PACKET + 4;
// Copter header
	net.out_buffer[0] = SYNC_CODE;
// packet type
	net.out_buffer[1] = PACKET_AUDIO;
	net.out_buffer[2] = 0;
	net.out_buffer[3] = 0;
	memcpy(net.out_buffer + 4, 
		sonar.samples + sonar.read_byte, 
		DEBUG2_PACKET);
	sonar.read_byte += DEBUG2_PACKET;
	if(sonar.read_byte >= DEBUG2_SIZE)
	{
		sonar.read_byte = 0;
		sonar.write_byte = 0;
	}
	
	sonar.audio_fragments++;
	if(sonar.audio_fragments < 1)
	{
		send_udp(payload_size, 
			send_audio_fragment,
			0);
	}
	else
	{
		send_udp(payload_size, 
			net_start_receive,
			0);
	}

}


#endif // DEBUG_SONAR2


#ifdef PIC_USE_MRF49XA
#ifndef LEFT_MODE


void send_radio_fragment()
{
//TRACE2
//print_number(copter_io.radio_size[!copter_io.current_radio_buffer]);
	int size = copter_io.radio_size[!copter_io.current_radio_buffer];
	if(size > 0)
	{
		int remainder = 64 - (size % 64);
		bzero(copter_io.radio_buffer[!copter_io.current_radio_buffer] + 
			size,
			remainder);
		size += remainder;

//TRACE2
//print_number(size);
#if defined(GROUND_MODE) || defined(LEFT_MODE)
		net_start_transmit(copter_io.radio_buffer[!copter_io.current_radio_buffer], 
			size, 
			net_start_receive);
#endif
	}
	else
	{
#ifdef GROUND_MODE
		net_start_receive();
#endif
	}
//TRACE2
}




#ifndef COPTER_MODE
void handle_xbee_receive()
{
/*
 * char buffer[1];
 * buffer[0] = radio.return_value;
 * send_uart(buffer, 1);
 */

	if(copter_io.radio_size[copter_io.current_radio_buffer] < RADIO_BUFFER_SIZE)
	{
		unsigned char *buffer = copter_io.radio_buffer[copter_io.current_radio_buffer];
		memcpy(buffer + copter_io.radio_size[copter_io.current_radio_buffer], 
			radio.receive_buf, 
			radio.receive_size);
		copter_io.radio_size[copter_io.current_radio_buffer] += radio.receive_size;
//send_uart(&radio.return_value, 1);
//		buffer[copter_io.radio_size[copter_io.current_radio_buffer]++] =
//			radio.return_value;
//		copter_io.packet_size -= radio.receive_size;
	}
}

// forward ground beacons
void handle_xbee_send()
{
	if(net.beacon_offset < COPTER_SERIAL_IN_SIZE)
	{
		radio.have_output = 1;
		radio.output_byte = net.beacon[net.beacon_offset++];
//net.beacon_offset = COPTER_SERIAL_IN_SIZE;
	}
}

#else // !COPTER_MODE


void handle_xbee_receive()
{
	handle_beacon();
//	print_buffer(radio.receive_buf, radio.receive_size);
}

// transmit telemetry
void handle_xbee_send()
{
	if(net.beacon_offset < radio.out_packet_size)
	{
		radio.have_output = 1;
		radio.output_byte = net.out_buffer[net.beacon_offset++];
//net.beacon_offset = COPTER_SERIAL_IN_SIZE;
	}
}






#endif // COPTER_MODE



#endif // !LEFT_MODE
#endif // PIC_USE_MRF49XA




#if defined(USE_FLOW) || (defined(USE_TCM8230) && !defined(COPTER_MODE))

void send_picture_fragment();

void next_picture_fragment()
{
	net.picture_fragments++;
	if(net.picture_fragments < PICTURE_FRAGMENTS)
	{
		send_picture_fragment();
	}
	else
	{
#if defined(PIC_USE_MRF49XA) && !defined(LEFT_MODE)
		send_radio_fragment();
#else
		net_start_receive();
#endif
	}
}










#if defined(USE_STEREO) && defined(DAISYCHAIN)


// move picture fragment from left eye SPI to right eye FIFO
void forward_picture_fragment()
{
//TRACE2
//print_number(net.left_usage);
//print_number(net.left_write);
//print_lf();

	if(net.left_usage <= LEFT_FIFO_SIZE - 
		PICTURE_FRAGMENT_SIZE - 
		SEGMENT_HEADER_SIZE)
	{
		memcpy(net.left_fifo + net.left_write,
			spi.dma_buffer + 4,
			PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE);
		net.left_usage += PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;
		net.left_write += PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;
		if(net.left_write >= LEFT_FIFO_SIZE)
		{
			net.left_write = 0;
		}
	}
}



void next_left_fragment()
{
	spi.time = TIM10->CNT;
	spi.current_function = picture_fragment_delay;
}




// send picture fragment from left eye
void send_picture_fragment_spi()
{
	if(camera.fifo_usage < PICTURE_FRAGMENT_SIZE)
	{
//TRACE2
// Only wait for beacon if configuration not received
		if(BIT_IS_CLEAR(copter.config_status, 0) ||
			BIT_IS_CLEAR(copter.config_status, 1) ||
			BIT_IS_CLEAR(copter.config_status, 2)) start_spi_receive();

		return;
	}

//TRACE2

// same picture fragment header as network
	int offset = 0;
	spi.dma_buffer[offset++] = 0;
	spi.dma_buffer[offset++] = 0;
	spi.dma_buffer[offset++] = 0;
	spi.dma_buffer[offset++] = 0;
	spi.dma_buffer[offset++] = SYNC_CODE;
	spi.dma_buffer[offset++] = PACKET_VIDEO;
// tag for left eye
	spi.dma_buffer[offset++] = 1;
	spi.dma_buffer[offset++] = 0;
// Store header
	*(u32*)(spi.dma_buffer + offset) = net.frame_offset;
	offset += 4;
	*(u32*)(spi.dma_buffer + offset) = net.frame_number;
	offset += 4;

//if(camera.fifo[camera.read_offset + 3] == 0xda) print_buffer(camera.fifo + camera.read_offset, 16);
	net.frame_offset++;
	memcpy(spi.dma_buffer + 4 + SEGMENT_HEADER_SIZE, 
		camera.fifo + camera.read_offset, 
		PICTURE_FRAGMENT_SIZE);

//print_buffer(spi.dma_buffer, 16);


	camera.read_offset += PICTURE_FRAGMENT_SIZE;
	if(camera.read_offset >= CAM_FIFO_SIZE)
		camera.read_offset = 0;
	camera.fifo_usage -= PICTURE_FRAGMENT_SIZE;
	start_spi_transmit(PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE);

	net.picture_fragments++;
// Nothing to receive after configuration beacon.
//	if(net.picture_fragments < LEFT_PICTURE_FRAGMENTS)
	if(1)
	{
		spi.next_function = next_left_fragment;
	}
	else
	{
		spi.next_function = start_spi_receive;
	}
}



#endif // USE_STEREO


#endif // USE_FLOW || (defined(USE_TCM8230) && !defined(COPTER_MODE))













#if defined(USE_FLOW) && defined(COPTER_MODE)


void send_picture_fragment()
{
	if(flow.preview_full)
	{
		int payload_size = PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;
		net.out_buffer[0] = SYNC_CODE;
		net.out_buffer[1] = PACKET_VIDEO;
		net.out_buffer[2] = 0;
		net.out_buffer[3] = 0;
		*(u32*)(net.out_buffer + 4) = flow.preview_offset;
		*(u32*)(net.out_buffer + 8) = flow.preview_number;
		memcpy(net.out_buffer + SEGMENT_HEADER_SIZE, 
			flow.preview_image + flow.preview_offset, 
			PICTURE_FRAGMENT_SIZE);
		flow.preview_offset += PICTURE_FRAGMENT_SIZE;
		if(flow.preview_offset >= flow.w * flow.h)
		{
			flow.preview_full = 0;
		}

		send_udp(payload_size, 
			next_picture_fragment,
			0);

	}
	else
	{
		net_start_receive();
	}
}


#endif // USE_FLOW


#ifdef USE_TCM8230

void send_picture_fragment()
{


// Test for no data left
	if(camera.fifo_usage < PICTURE_FRAGMENT_SIZE 
#ifdef USE_STEREO
		&&
		net.left_usage < PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE
#endif

		)
	{
//TRACE2

#if defined(PIC_USE_MRF49XA) && !defined(LEFT_MODE)
		send_radio_fragment();
#else
		net_start_receive();
#endif
		return;
	}



// Have to memcpy from the core coupled memory to the network memory
// for USB to work.
	int payload_size = PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;


// send left eye packet if no right eye data
#ifdef USE_STEREO
	if(camera.fifo_usage < PICTURE_FRAGMENT_SIZE)
	{
//print_buffer(net.left_fifo + net.left_read, 8);
		memcpy(net.out_buffer, 
			net.left_fifo + net.left_read, 
			PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE);
		net.left_usage -= PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;
		net.left_read += PICTURE_FRAGMENT_SIZE + SEGMENT_HEADER_SIZE;
		if(net.left_read >= LEFT_FIFO_SIZE)
			net.left_read = 0;
		
	#ifndef USE_WIFI
		net_start_transmit(net.out_buffer, 
			payload_size, 
			next_picture_fragment);
	#else
		send_udp(payload_size, 
			next_picture_fragment,
			0);
	#endif
	
	
		return;
	}
#endif // USE_STEREO


// Copter header
	net.out_buffer[0] = SYNC_CODE;
// packet type
	net.out_buffer[1] = PACKET_VIDEO;
	net.out_buffer[2] = 0;
	net.out_buffer[3] = 0;


// Store header
	*(u32*)(net.out_buffer + 4) = net.frame_offset;
	*(u32*)(net.out_buffer + 8) = net.frame_number;
	

// Store picture data
// Must start on a multiple of 4 for memcpy to go fast

	memcpy(net.out_buffer + SEGMENT_HEADER_SIZE, 
		camera.fifo + camera.read_offset, 
		PICTURE_FRAGMENT_SIZE);

//TRACE2
//print_number(camera.fifo_usage);
// Update offsets after resending
	net.resends_done++;
	if(net.resends_done >= TOTAL_RESENDS)
	{
// increments forever
		net.frame_offset++;
		camera.read_offset += PICTURE_FRAGMENT_SIZE;
		if(camera.read_offset >= CAM_FIFO_SIZE)
			camera.read_offset = 0;
		camera.fifo_usage -= PICTURE_FRAGMENT_SIZE;
		net.resends_done = 0;
	}







// send fragment
#if defined(LEFT_MODE) && defined(DAISYCHAIN)
// send picture fragment to SPI
	write_spi(net.out_buffer, payload_size);
#else
// send to network
	#ifndef USE_WIFI
	net_start_transmit(net.out_buffer, 
		payload_size, 
		next_picture_fragment);
	#else
	send_udp(payload_size, 
		next_picture_fragment,
		0);
	#endif
	
#endif // GROUND_MODE


}


#endif // USE_TCM8230



#ifndef MARCY2






void send_telemetry()
{
// Return telemetry packets
	int offset = 0;
	int i;




#if !defined(GROUND_MODE) || defined(ROVER) || defined(COPTER_MODE)
// If the aircraft sends configuration packets, only send those.
// this is to test the camera
// send config status
	if((BIT_IS_SET(copter.status1, CONFIG_VALID) ||
		BIT_IS_SET(copter.status1, FORCE_CONFIG_SEND)) &&
		offset + CONFIG_PACKET_SIZE <= TELEMETRY_DATA_SIZE)
	{





		int start = offset;
		CLEAR_BIT_(copter.status1, CONFIG_VALID)
		net.out_buffer[offset++] = SYNC_CODE;
// packet type. packet count is not needed for config
		net.out_buffer[offset++] = PACKET_CONFIG;
		net.out_buffer[offset++] = copter.config_status;
// padding
		net.out_buffer[offset++] = 0;
		

		while(offset - start < CONFIG_PACKET_SIZE - 2)
		{
			net.out_buffer[offset++] = 0;
		}

// CRC ignored.  Packets are aligned in the network packets.
		copter_chksum(net.out_buffer + start + CONFIG_PACKET_SIZE - 2, 
			net.out_buffer + start, 
			CONFIG_PACKET_SIZE - 2);
		offset += 2;
	}



	if(BIT_IS_SET(copter.status2, HAVE_BATTERY)
#ifdef PIC_USE_IMU
	 	&& BIT_IS_SET(copter.status2, HAVE_IMU)
#endif
	)
	{
//TRACE2
//print_number(copter.current_battery);
		CLEAR_BIT_(copter.status2, HAVE_BATTERY)
		CLEAR_BIT_(copter.status2, HAVE_IMU)

#ifdef PIC_USE_IMU
		for(i = 0; 
			i < copter.analog_history &&
				offset + ANALOG_PACKET_SIZE <= TELEMETRY_DATA_SIZE; 
			i++)
#endif
		{
			int start = offset;
			net.out_buffer[offset++] = SYNC_CODE;
	// packet type & count
			net.out_buffer[offset++] = PACKET_ANALOG | (copter.analog_count & 0xf);
			copter.analog_count++;

#ifdef PIC_USE_IMU
			net.out_buffer[offset++] = copter.accel_x[i] & 0xff;
			net.out_buffer[offset++] = (copter.accel_x[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.accel_y[i] & 0xff;
			net.out_buffer[offset++] = (copter.accel_y[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.accel_z[i] & 0xff;
			net.out_buffer[offset++] = (copter.accel_z[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.gyro_x[i] & 0xff;
			net.out_buffer[offset++] = (copter.gyro_x[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.gyro_y[i] & 0xff;
			net.out_buffer[offset++] = (copter.gyro_y[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.gyro_z[i] & 0xff;
			net.out_buffer[offset++] = (copter.gyro_z[i] >> 8) & 0xff;
#endif

	// send the battery level
			net.out_buffer[offset++] = copter.current_battery & 0xff;
			net.out_buffer[offset++] = (copter.current_battery >> 8) & 0xff;



	// CRC ignored.  Packets are aligned in the network packets.
			while(offset - start < ANALOG_PACKET_SIZE - 2)
			{
				net.out_buffer[offset++] = 0;
			}

// CRC ignored.  Packets are aligned in the network packets.
			copter_chksum(net.out_buffer + start + ANALOG_PACKET_SIZE - 2, 
				net.out_buffer + start, 
				ANALOG_PACKET_SIZE - 2);
			offset += 2;
//TRACE2
//print_number(copter.current_battery);

		}

		
		copter.analog_history = 0;
	}

#endif // !defined(GROUND_MODE) || defined(ROVER) || defined(COPTER_MODE)







#ifdef GROUND_MODE


// buttons
#ifdef POSITION_TRAINING
	int start = offset;
	net.out_buffer[offset++] = SYNC_CODE;
	net.out_buffer[offset++] = PACKET_BUTTONS;
	net.out_buffer[offset++] = copter.start_button |
		(copter.reset_button << 1) |
		(copter.config_button << 2);
	net.out_buffer[offset++] = 0;
	copter_chksum(net.out_buffer + start + BUTTON_PACKET_SIZE - 2, 
			net.out_buffer + start, 
			BUTTON_PACKET_SIZE - 2);
	offset += 2;
//	TRACE2
//	print_number(offset);
#endif // POSITION_TRAINING



// Failed use of an IMU for camera tracking
#ifdef USE_GROUND_IMU

	if(imu.have_imu)
	{
		imu.have_imu = 0;
		int start = offset;
		net.out_buffer[offset++] = SYNC_CODE;
		net.out_buffer[offset++] = PACKET_IMU | (imu.packet_count & 0xf);
		imu.packet_count++;

/*
 * 		net.out_buffer[offset++] = imu.mag_x & 0xff;
 * 		net.out_buffer[offset++] = (imu.mag_x >> 8) & 0xff;
 * 		net.out_buffer[offset++] = imu.mag_y & 0xff;
 * 		net.out_buffer[offset++] = (imu.mag_y >> 8) & 0xff;
 * 		net.out_buffer[offset++] = imu.mag_z & 0xff;
 * 		net.out_buffer[offset++] = (imu.mag_z >> 8) & 0xff;
 * 		net.out_buffer[offset++] = imu.accel_x & 0xff;
 * 		net.out_buffer[offset++] = (imu.accel_x >> 8) & 0xff;
 * 		net.out_buffer[offset++] = imu.accel_y & 0xff;
 * 		net.out_buffer[offset++] = (imu.accel_y >> 8) & 0xff;
 * 		net.out_buffer[offset++] = imu.accel_z & 0xff;
 * 		net.out_buffer[offset++] = (imu.accel_z >> 8) & 0xff;
 */

  		net.out_buffer[offset++] = imu.analog0 & 0xff;
  		net.out_buffer[offset++] = (imu.analog0 >> 8) & 0xff;
  		net.out_buffer[offset++] = imu.analog1 & 0xff;
  		net.out_buffer[offset++] = (imu.analog1 >> 8) & 0xff;

// extra space for gyros
		while(offset - start < IMU_PACKET_SIZE - 2)
		{
			net.out_buffer[offset++] = 0;
		}
		
		copter_chksum(net.out_buffer + start + IMU_PACKET_SIZE - 2, 
				net.out_buffer + start, 
				IMU_PACKET_SIZE - 2);
		offset += 2;
	}


#endif // USE_GROUND_IMU
#endif // GROUND_MODE






#ifdef USE_SONAR2
	if(copter.sonar_history > 0 &&
		offset + SONAR_PACKET_SIZE <= TELEMETRY_DATA_SIZE)
	{
		for(i = 0; i < copter.sonar_history; i++)
		{
			int start = offset;
			net.out_buffer[offset++] = SYNC_CODE;
	// packet type & count
			net.out_buffer[offset++] = PACKET_SONAR | (copter.sonar_count & 0xf);
			copter.sonar_count++;

			net.out_buffer[offset++] = copter.sonar_channel[i];
			net.out_buffer[offset++] = 0;
			net.out_buffer[offset++] = copter.sonar_delay[i] & 0xff;
			net.out_buffer[offset++] = (copter.sonar_delay[i] >> 8) & 0xff;

			while(offset - start < SONAR_PACKET_SIZE)
			{
				net.out_buffer[offset++] = 0;
			}
		}
		
		copter.sonar_history = 0;
	}
#endif // USE_SONAR2







#ifdef ROVER
// wheel encoders
	net.out_buffer[offset++] = SYNC_CODE;
// packet type & count
	net.out_buffer[offset++] = PACKET_ROVER;
	for(i = 0; i < 4; i++)
	{
		net.out_buffer[offset++] = copter.encoder[i] & 0xff;
		net.out_buffer[offset++] = (copter.encoder[i] >> 8) & 0xff;
	}

	net.out_buffer[offset++] = copter.finished_rover_command;
	net.out_buffer[offset++] = 0;
// ignored CRC
	net.out_buffer[offset++] = 0;
	net.out_buffer[offset++] = 0;
	



	if(BIT_IS_SET(copter.status1, MAG_VALID))
	{
		CLEAR_BIT_(copter.status1, MAG_VALID)
		
		
/*
 * TRACE2
 * print_number(copter.mag_history);
 */
		for(i = 0; 
			i < copter.mag_history &&
				offset + MAG_PACKET_SIZE <= TELEMETRY_DATA_SIZE; 
			i++)
		{
			int start = offset;
			net.out_buffer[offset++] = SYNC_CODE;
	// packet type & count
			net.out_buffer[offset++] = PACKET_MAG | (copter.mag_count & 0x0f);
			copter.mag_count++;

/*
 * print_number(copter.mag_x[i]);
 * print_number(copter.mag_y[i]);
 * print_number(copter.mag_z[i]);
 * print_lf();
 */
			net.out_buffer[offset++] = copter.mag_x[i] & 0xff;
			net.out_buffer[offset++] = (copter.mag_x[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.mag_y[i] & 0xff;
			net.out_buffer[offset++] = (copter.mag_y[i] >> 8) & 0xff;
			net.out_buffer[offset++] = copter.mag_z[i] & 0xff;
			net.out_buffer[offset++] = (copter.mag_z[i] >> 8) & 0xff;
	// CRC ignored.  Packets are aligned in the network packets.
			while(offset - start < MAG_PACKET_SIZE)
			{
				net.out_buffer[offset++] = 0;
			}
		}
		
		copter.mag_history = 0;
	}

#endif // ROVER








#ifdef COPTER_MODE
#ifdef USE_AZIMUTH
// Send azimuth data.  Must be last because it goes to the end.
	while(copter.azimuth_buffer_usage > AZIMUTH_SEND_SAMPLES &&
		offset + AZIMUTH_PACKET_SIZE <= TELEMETRY_DATA_SIZE)
	{
		int start = offset;
		net.out_buffer[offset++] = SYNC_CODE;
// packet type. packet count is needed for azimuth
		net.out_buffer[offset++] = PACKET_AZIMUTH | 
			(copter.azimuth_count & 0x0f);
		copter.azimuth_count++;
// store detected period
		net.out_buffer[offset++] = copter.azimuth_period & 0xff;
		net.out_buffer[offset++] = (copter.azimuth_period >> 8) & 0xff;
// store total readings without the current buffer
		int azimuth_total = copter.azimuth_total - copter.azimuth_buffer_usage;
		net.out_buffer[offset++] = azimuth_total & 0xff;
		net.out_buffer[offset++] = (azimuth_total >> 8) & 0xff;
		memcpy(net.out_buffer + offset,
			copter.azimuth_buffer + copter.azimuth_read_ptr,
			AZIMUTH_SEND_SAMPLES * AZIMUTH_SAMPLE_BYTES);
		offset += AZIMUTH_SEND_SAMPLES * AZIMUTH_SAMPLE_BYTES;
		copter.azimuth_read_ptr += AZIMUTH_SEND_SAMPLES * AZIMUTH_SAMPLE_BYTES;
		copter.azimuth_buffer_usage -= AZIMUTH_SEND_SAMPLES;
		if(copter.azimuth_read_ptr >= AZIMUTH_BUFFER_SIZE)
		{
			copter.azimuth_read_ptr = 0;
		}
// CRC ignored.  Packets are aligned in the network packets.
		while(offset - start < AZIMUTH_PACKET_SIZE)
		{
			net.out_buffer[offset++] = 0;
		}
	}



#endif // !USE_AZIMUTH



#endif // COPTER_MODE




//TRACE2
//print_text("offset=");
//print_number(offset);


// Have data to send
	if(offset > 0)
	{
//TRACE2

// Pad to TELEMETRY_DATA_SIZE to avoid an unknown size related bug
		while(offset < TELEMETRY_DATA_SIZE)
			net.out_buffer[offset++] = 0;



#ifndef USE_WIFI

#if defined(GROUND_MODE) || defined(LEFT_MODE)
	net_start_transmit(net.out_buffer, 
		offset, 
	#ifdef USE_TCM8230
		send_picture_fragment
	#else
		#ifdef PIC_USE_MRF49XA
			send_radio_fragment
		#else
			net_start_receive
		#endif
	#endif
	);
#endif

#else // ! USE_WIFI


//print_buffer(net.out_buffer, offset);
		send_udp(offset, 
	#ifdef USE_TCM8230
			send_picture_fragment,
	#else
		#ifdef PIC_USE_MRF49XA
			send_radio_fragment,
		#else
			net_start_receive,
		#endif
	#endif
			0);


#endif // USE_WIFI

	}
	else
	{
// Nothing to send
#if defined(USE_TCM8230) && !defined(COPTER_MODE)
 		send_picture_fragment();
#else
	#ifdef PIC_USE_MRF49XA
		send_radio_fragment();
	#else
		net_start_receive();
	#endif
#endif
	}
}


int handle_config_section(int section_id, unsigned char *data, int offset)
{
	switch(section_id)
	{
		case 0:
// 6
#if defined(USE_NAV) && defined(COPTER_MODE)
			nav.gyro_x_axis = data[offset++];
			nav.gyro_y_axis = data[offset++];
			nav.gyro_z_axis = data[offset++];
			nav.gyro_x_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.gyro_y_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.gyro_z_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.mag_x_axis = data[offset++];
			nav.mag_y_axis = data[offset++];
			nav.mag_z_axis = data[offset++];
			nav.mag_x_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.mag_y_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.mag_z_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.accel_x_axis = data[offset++];
			nav.accel_y_axis = data[offset++];
			nav.accel_z_axis = data[offset++];
			nav.accel_x_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.accel_y_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.accel_z_sign = data[offset++] & 0x80 ? -1 : 1;
			nav.angle_to_gyro = data[offset] | 
				(data[offset + 1] << 8) | 
				(data[offset + 2] << 16) | 
				(data[offset + 3] << 24);
			offset += 4;
#endif // defined(USE_NAV) && defined(COPTER_MODE)
			break;

		case 1:
// 6
/*
 * 			nav.mag_x_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.mag_y_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.mag_z_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.mag_x_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.mag_y_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.mag_z_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 
 * 
 * 			nav.attitude_blend = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 * 			nav.bias_blend = (int16_t)(data[offset] | (data[offset + 1] << 8));
 * 			offset += 2;
 */


// PWM range
			copter.min_throttle = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.max_throttle = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.throttle_input_pwm[0] = copter.min_throttle;

			unsigned char reverse_flags = data[offset++];
			pwm.reverse_throttle = BIT_IS_SET(reverse_flags, 0);
			pwm.reverse_yaw = BIT_IS_SET(reverse_flags, 1);
			pwm.reverse_roll = BIT_IS_SET(reverse_flags, 2);
			pwm.reverse_pitch = BIT_IS_SET(reverse_flags, 3);

/*
 * 			nav.gyro_bandwidth = data[offset++];
 * 			nav.accel_bandwidth = data[offset++];
 * 			nav.mag_bandwidth = data[offset++];
 */


TRACE2
print_text("min_throttle=");
print_number(copter.min_throttle);
print_text("max_throttle=");
print_number(copter.max_throttle);
print_text("reverse_throttle=");
print_number(pwm.reverse_throttle);
print_text("reverse_yaw=");
print_number(pwm.reverse_yaw);
print_text("reverse_roll=");
print_number(pwm.reverse_roll);
print_text("reverse_pitch=");
print_number(pwm.reverse_pitch);

			break;



		case 2:
#ifdef USE_TCM8230
// width / 8
			camera.frame_w = data[offset++] * 8;
// height / 8
			camera.frame_h = data[offset++] * 8;

			camera.half_width = BIT_IS_SET(data[offset], 0);
			camera.use_greyscale = BIT_IS_SET(data[offset], 1);
			camera.use_color = BIT_IS_SET(data[offset], 2);
			camera.use_chroma = BIT_IS_SET(data[offset], 3);
			camera.use_threshold = BIT_IS_SET(data[offset], 4);
			offset++;


			camera.shutter_speed_code = data[offset++];
			camera.clockspeed_code = data[offset++];
			camera.downsample_x = data[offset++];
			camera.downsample_y = data[offset++];
			camera.y_threshold = data[offset++];
			camera.u_threshold = data[offset++];
			camera.v_threshold = data[offset++];

			set_camera_clockspeed();

//flush_uart();
//print_cam_config();
			camera.have_config = 1;
#endif // USE_TCM8230


#ifdef ROVER
			gyro.gyro_to_angle = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			reset_gyro();
#endif // ROVER


			break;

		case 3:
			copter.servo_center = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.servo_mag = data[offset] | (data[offset + 1] << 8);
			offset += 2;

			copter.min_servo = copter.servo_center - copter.servo_mag;
			copter.max_servo = copter.servo_center + copter.servo_mag;
			copter.servo_input_pwm[0] = copter.servo_center;
			copter.servo_input_pwm[1] = copter.servo_center;
			copter.servo_input_pwm[2] = copter.servo_center;

#if defined(USE_NAV) && defined(COPTER_MODE)
			nav.compass_offset = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			nav.compass_sign = data[offset++] & 0x80 ? -1 : 1;
#endif // defined(USE_NAV) && defined(COPTER_MODE)

TRACE2
print_text("min_servo=");
print_number(copter.min_servo);
print_text("max_servo=");
print_number(copter.max_servo);

			break;
	}
	
	return offset;
}


#endif // !MARCY2






#if defined(MARCY1) || defined(MARCY3)

// These have a USB ground station connected to a PC

void handle_beacon()
{
	if(!usb.have_beacon) return;

	unsigned char *data = usb.in_buf;

// switch radio buffer when it is known to not be getting filled
	copter_io.current_radio_buffer = !copter_io.current_radio_buffer;
	copter_io.radio_size[copter_io.current_radio_buffer] = 0;


// forward beacon to radio
#ifndef LEFT_MODE
	radio.out_packet_size = COPTER_SERIAL_IN_SIZE;
	memcpy(net.beacon, data, COPTER_SERIAL_IN_SIZE);
	net.beacon_offset = 0;
#endif

	int offset = 0;
	int i;


	if(data[offset] == SYNC_CODE)
	{
// prepare image routine
		net.picture_fragments = 0;

#ifdef DEBUG_SONAR2
		sonar.audio_fragments = 0;
#endif

// sync code
		offset++;
		
		unsigned char ground_status = data[offset++];
//TRACE2
//print_hex(ground_status);
//TRACE2
//print_number(camera.fifo_usage);

		copter.shutdown_timeout = copter.shutdown_timeout_cycles;


// counter
		copter_io.packet_counter = data[offset++];
// POV panel/padding
		offset++;

// offset = 4

		if(BIT_IS_SET(ground_status, CONFIG_VALID_BIT))
		{
			int section_id = data[offset++];
			offset++;
TRACE2
print_text("config ");
print_number(section_id);
			offset = handle_config_section(section_id, data, offset);

		}
// Not config. Not used by the ground except for RC transmitter hacks.
		else
		{
			COPY_BIT(copter.status1, 
				GROUND_ENGINE_ON, 
				ground_status, 
				ENGINE_MODE_BIT);




			COPY_BIT(copter.status2,
				AUTOPILOT_ON,
				ground_status,
				AUTOPILOT_ENABLE_BIT);



// PWM
			SET_BIT_(copter.status1, HAVE_PWM);

// Capture turret PWM


			copter.turret_pwm0 = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.turret_pwm1 = data[offset] | (data[offset + 1] << 8);
			offset += 2;


#if (SERVOS > 0)
			for(i = 0; i < SERVOS; i++)
			{
				copter.servo_input_pwm[i] = data[offset] | (data[offset + 1] << 8);
				offset += 2;
			}
#endif

#if (ROTORS > 0)
			for(i = 0; i < ROTORS; i++)
			{
				copter.throttle_input_pwm[i] = data[offset] | (data[offset + 1] << 8);
				offset += 2;
			}
#endif





/*
 * TRACE2
 * print_number(ENGINE_ON);
 * print_number(copter.throttle_input_pwm[0]);
 * print_number(copter.servo_input_pwm[0]);
 * print_number(copter.servo_input_pwm[1]);
 * print_number(copter.servo_input_pwm[2]);
 */



// TRACE2
// print_number(offset);
// print_number(copter.turret_pwm0);
// print_number(copter.turret_pwm1);

#ifndef LEFT_MODE
			update_turret();
#endif


/*
 * TRACE2
 * print_number(copter.throttle_input_pwm[0]);
 * print_number(copter.servo_input_pwm[0]);
 * print_number(copter.servo_input_pwm[1]);
 * print_number(copter.servo_input_pwm[2]);
 */






		}


	}




	send_telemetry();



}



#else  // defined(MARCY1) || defined(MARCY3)




void handle_beacon()
{
#ifndef USE_WIFI

#ifdef COPTER_MODE
#if defined(PIC_USE_MRF49XA_GROUND) || defined(PIC_USE_MRF49XA)
	unsigned char *data = radio.receive_buf;
#endif

#ifdef USE_BLUETOOTH
	unsigned char *data = bluetooth.receive_buf;
#endif

	unsigned char chksum[2];

	if(data[0] != SYNC_CODE) return;

	copter_chksum(chksum, data, COPTER_SERIAL_IN_SIZE - 2);
	if(chksum[0] != data[COPTER_SERIAL_IN_SIZE - 2] ||
		chksum[1] != data[COPTER_SERIAL_IN_SIZE - 1])
	{
		return;
	}

	
	
#else // COPTER_MODE
	if(!usb.have_beacon) return;

	unsigned char *data = usb.in_buf;
#endif // !COPTER_MODE

#else // !USE_WIFI

	unsigned char *buf = net.in_buffer + PACKET_OFFSET;
	unsigned char *ip_buf = buf + 32;
	unsigned char *udp_buf = ip_buf + 20;
	unsigned char *data = udp_buf + 8;

#endif  // USE_WIFI



#ifndef COPTER_MODE
#ifdef PIC_USE_MRF49XA

//TRACE2
//print_buffer(data, COPTER_SERIAL_IN_SIZE);

// switch radio buffer when it is known to not be getting filled
	copter_io.current_radio_buffer = !copter_io.current_radio_buffer;
	copter_io.radio_size[copter_io.current_radio_buffer] = 0;


// forward beacon to radio
	radio.out_packet_size = COPTER_SERIAL_IN_SIZE;
	memcpy(net.beacon, data, COPTER_SERIAL_IN_SIZE);
	net.beacon_offset = 0;
#endif // PIC_USE_MRF49XA
#endif // !COPTER_MODE

	int offset = 0;
	int i;


	if(data[offset] == SYNC_CODE)
	{
// prepare image routine
		net.picture_fragments = 0;

#ifdef DEBUG_SONAR2
		sonar.audio_fragments = 0;
#endif

// sync code
		offset++;
		
		unsigned char ground_status = data[offset++];


		if(BIT_IS_SET(ground_status, RESET_BIT))
		{
			SET_BIT_(copter.status1, FORCE_CONFIG_SEND);
			copter.config_status = 0;
		}


//TRACE2
//print_hex(ground_status);
		
#ifdef ROVER
		if(
// Reset shutdown timer if configuration & radio received
			BIT_IS_SET(copter.config_status, 0) &&
			BIT_IS_SET(copter.config_status, 1) &&
			BIT_IS_SET(copter.config_status, 2) &&
			BIT_IS_SET(copter.config_status, 3) &&
			BIT_IS_SET(ground_status, RADIO_VALID_BIT))
#endif // ROVER
		{
			copter.shutdown_timeout = copter.shutdown_timeout_cycles;
		}


// counter
		copter_io.packet_counter = data[offset++];
// POV panel/padding
		offset++;

// offset = 4

		if(BIT_IS_SET(ground_status, CONFIG_VALID_BIT))
		{
			int section_id = data[offset++];
TRACE2
print_text("config ");
print_number(section_id);
#if defined(ROVER) || defined(MARCY2)
			CLEAR_BIT_(copter.status1, FORCE_CONFIG_SEND);
			CLEAR_BIT_(copter.status1, GROUND_ENGINE_ON);
			CLEAR_BIT_(copter.status2, AUTOPILOT_ON);
			CLEAR_BIT_(copter.status2, COPTER_ENGINE_ON);
// flag to send confirmation of configuration received
			SET_BIT_(copter.status1, CONFIG_VALID);



// padding
			offset++;
// offset = 6
// ground computer reset if control packet previously received
			if(copter.got_control)
			{
				copter.got_control = 0;
				copter.config_status = 0;
				copter.shutdown_timeout = 0;
			}
			
			SET_BIT_(copter.config_status, section_id)
			offset = handle_config_section(section_id, data, offset);

#endif // defined(ROVER) || defined(MARCY2)

		}
// Not config. Not used by the ground except for RC transmitter hacks.
		else
		{
			COPY_BIT(copter.status1, 
				GROUND_ENGINE_ON, 
				ground_status, 
#ifdef ROVER
				ENGINE_ON_BIT);
#else
				ENGINE_MODE_BIT);

#endif




			COPY_BIT(copter.status2,
				AUTOPILOT_ON,
				ground_status,
				AUTOPILOT_ENABLE_BIT);

#ifdef POSITION_TRAINING
			if(!BIT_IS_SET(ground_status, RECORDING_BIT) &&
				autopilot.recording_function)
			{
				stop_recording();
			}

			if(ENGINE_ON &&
				BIT_IS_SET(copter.status2, IS_RECORDING))
			{
				stop_recording();
			}

			if(BIT_IS_SET(ground_status, RECORDING_BIT) &&
				!autopilot.recording_function &&
				!ENGINE_ON)
			{
// start recording
				
				start_recording();
			}
#endif // POSITION_TRAINING


#if defined(USE_NAV) && defined(COPTER_MODE)
			if(BIT_IS_SET(ground_status, DO_LANDING_BIT) &&
				ENGINE_ON &&
				autopilot.landing_state == LANDING_INACTIVE)
			{
				autopilot.landing_state = LANDING_ACTIVE;
			}


// Mag calibration
			if(BIT_IS_SET(ground_status, CALIBRATE_BIT) &&
				!BIT_IS_SET(copter.status2, COPTER_CALIBRATE_MAG) &&
				!ENGINE_ON)
			{
				SET_BIT_(copter.status2, COPTER_CALIBRATE_MAG);
				
				nav.mag_x_max = -0xffff;
				nav.mag_y_max = -0xffff;
				nav.mag_z_max = -0xffff;
				nav.mag_x_min = 0xffff;
				nav.mag_y_min = 0xffff;
				nav.mag_z_min = 0xffff;
				nav.compass_offset = 0;
			}
			else
			if(!BIT_IS_SET(ground_status, CALIBRATE_BIT) &&
				BIT_IS_SET(copter.status2, COPTER_CALIBRATE_MAG))
			{
				CLEAR_BIT_(copter.status2, COPTER_CALIBRATE_MAG);
				nav.compass_offset = -nav.current_heading;
				save_settings();
			}
#endif // defined(USE_NAV) && defined(COPTER_MODE)

// PWM
			SET_BIT_(copter.status1, HAVE_PWM);



#ifdef MARCY2
			copter.throttle_input_pwm[0] = READ_UINT16(data, offset);

			copter.roll_input = READ_FLOAT32(data, offset);
			copter.pitch_input = READ_FLOAT32(data, offset);
			copter.heading_input = READ_FLOAT32(data, offset);
//TRACE2
//print_float(copter.roll_input);
//print_float(copter.pitch_input);
//print_float(copter.heading_input);

			copter.manual_alt = READ_FLOAT16(data, offset);
			copter.manual_yaw = READ_FLOAT16(data, offset);
			copter.manual_x = READ_FLOAT16(data, offset);
			copter.manual_y = READ_FLOAT16(data, offset);


/*
 * TRACE2
 * print_float(copter.manual_alt);
 * print_float(copter.manual_yaw);
 * print_float(copter.manual_x);
 * print_float(copter.manual_y);
 */

#else // MARCY2


#if (SERVOS > 0)
			for(i = 0; i < SERVOS; i++)
			{
				copter.servo_input_pwm[i] = data[offset] | (data[offset + 1] << 8);
				offset += 2;
			}
#endif

#if (ROTORS > 0)
			for(i = 0; i < ROTORS; i++)
			{
				copter.throttle_input_pwm[i] = data[offset] | (data[offset + 1] << 8);
				offset += 2;
			}
#endif



#endif // !MARCY2


// Rover packet
#ifdef ROVER
// camera enable
			if(BIT_IS_SET(ground_status, DISABLE_CAMERA_BIT))
			{
				copter.disable_camera = 1;
			}
			else
			{
				copter.disable_camera = 0;
			}

// servo PWM
			for(i = 0; i < 4; i++)
			{
				copter.pwm_input[i] = data[offset] | (data[offset + 1] << 8);
				offset += 2;
			}

// command
			copter.rover_command = data[offset++];
			copter.rover_command_number = data[offset++];
			copter.target_heading = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_x_axis = data[offset++];
			copter.mag_y_axis = data[offset++];
			copter.mag_z_axis = data[offset++];
			copter.mag_x_sign = (data[offset++] & 0x80) ? -1 : 1;
			copter.mag_y_sign = (data[offset++] & 0x80) ? -1 : 1;
			copter.mag_z_sign = (data[offset++] & 0x80) ? -1 : 1;
			copter.mag_x_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_y_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_z_max = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_x_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_y_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			copter.mag_z_min = (int16_t)(data[offset] | (data[offset + 1] << 8));
			offset += 2;
			
			
/*
 * 			TRACE2
 * 			print_number(copter.target_heading);
 * 			print_number(copter.mag_x_axis);
 * 			print_number(copter.mag_y_axis);
 * 			print_number(copter.mag_z_axis);
 * 			print_number(copter.mag_x_sign);
 * 			print_number(copter.mag_y_sign);
 * 			print_number(copter.mag_z_sign);
 * 			print_number(copter.mag_x_max);
 * 			print_number(copter.mag_y_max);
 * 			print_number(copter.mag_z_max);
 * 			print_number(copter.mag_x_min);
 * 			print_number(copter.mag_y_min);
 * 			print_number(copter.mag_z_min);
 */
			
			
#endif // ROVER

//TRACE2
//print_number(copter.throttle0_input_pwm);


#ifdef USE_AZIMUTH
// Azimuth
			copter.azimuth_threshold = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.azimuth_on = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.azimuth_off = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.azimuth_level = data[offset] | (data[offset + 1] << 8);
			offset += 2;
#endif // USE_AZIMUTH




// Capture turret PWM
//TRACE2
#if defined(GROUND_MODE)


			copter.turret_pwm0 = data[offset] | (data[offset + 1] << 8);
			offset += 2;
			copter.turret_pwm1 = data[offset] | (data[offset + 1] << 8);
			offset += 2;

// TRACE2
// print_number(offset);
// print_number(copter.turret_pwm0);
// print_number(copter.turret_pwm1);

			update_turret();
#endif // GROUND_MODE



/*
 * TRACE2
 * print_number(copter.throttle_input_pwm[0]);
 * print_number(copter.servo_input_pwm[0]);
 * print_number(copter.servo_input_pwm[1]);
 * print_number(copter.servo_input_pwm[2]);
 */






		}


#if defined(GROUND_MODE) && defined(USE_TCM8230) && defined(USE_STEREO) && defined(DAISYCHAIN)
// only send configuration to left eye, to avoid colliding with SPI video packets
		if(BIT_IS_SET(ground_status, CONFIG_VALID_BIT)) 
			write_spi(data, COPTER_SERIAL_IN_SIZE);
#endif  // GROUND_MODE



	}


// no telemetry for ground components on SPI
#if defined(LEFT_MODE) && defined(DAISYCHAIN)
	return;
#endif







	send_telemetry();
//TRACE2








}


#endif // !(defined(MARCY1) || defined(MARCY3))




#else // !defined(PACKET_FORWARD) || defined(COPTER_MODE)

// packet forwarding only
void handle_beacon()
{
	unsigned char *buf = net.in_buffer + PACKET_OFFSET;
	unsigned char *ip_buf = buf + 32;
	unsigned char *udp_buf = ip_buf + 20;
	unsigned char *data = udp_buf + 8;

// switch radio buffer when it is known to not be getting filled
	copter_io.current_radio_buffer = !copter_io.current_radio_buffer;
	copter_io.radio_size[copter_io.current_radio_buffer] = 0;


// forward telemetry from radio to wifi
	int size = copter_io.radio_size[!copter_io.current_radio_buffer];
	if(size > 0)
	{
		int remainder = 64 - (size % 64);
		bzero(copter_io.radio_buffer[!copter_io.current_radio_buffer] + 
			size,
			remainder);
		size += remainder;
		net_start_transmit(copter_io.radio_buffer[!copter_io.current_radio_buffer], 
			size, 
			net_start_receive);
	}
	else
	{
		net_start_receive();
	}




// forward beacon from wifi to radio
	radio.out_packet_size = COPTER_SERIAL_IN_SIZE;
	memcpy(net.beacon, data, COPTER_SERIAL_IN_SIZE);
	net.beacon_offset = 0;
}

void handle_xbee_receive()
{

// append packet to current radio buffer
	if(copter_io.radio_size[copter_io.current_radio_buffer] < RADIO_BUFFER_SIZE)
	{
		unsigned char *buffer = copter_io.radio_buffer[copter_io.current_radio_buffer];
		memcpy(buffer + copter_io.radio_size[copter_io.current_radio_buffer], 
			radio.receive_buf, 
			radio.receive_size);
		copter_io.radio_size[copter_io.current_radio_buffer] += radio.receive_size;
	}
}

// forward ground beacons
void handle_xbee_send()
{
	if(net.beacon_offset < COPTER_SERIAL_IN_SIZE)
	{
		radio.have_output = 1;
		radio.output_byte = net.beacon[net.beacon_offset++];
//net.beacon_offset = COPTER_SERIAL_IN_SIZE;
	}
}


#endif // defined(PACKET_FORWARD) && !defined(COPTER_MODE)






void init_copter_io()
{

	bzero(&copter_io, sizeof(copter_io_t));


#ifdef USE_WIFI
	string_end = string + TEXTLEN;
	new_channel = copter.channel;
	strcpy(new_ssid, copter.ssid);
#endif

}












