/*
 * STM32F4 Wifi flight controller
 * Copyright (C) 2012-2013 Adam Williams <broadcast at earthling dot net>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include "linux.h"
#include "wifi.h"

#include "camera.h"
#include "copter.h"
#include "copter_io.h"
#include "core.h"
#include "net.h"
#include "rtl_usb.h"
#include "uart.h"
#include "usb.h"

#include "arm/config.h"
#include "pic/config.h"


net_t net;

#ifdef USE_WIFI


const unsigned char cname_text[] = MARCY2_CNAME;
const unsigned char domain_text[] = MARCY2_DOMAIN;

const unsigned char ground_mac[] = GROUND_MAC;
const unsigned char ground_ip[] = GROUND_IP;
const unsigned char copter_ip[] = MARCY2_IP;

// Ground & copter have same IP address
const unsigned char this_ip[] = MARCY2_IP;

// Show network transactions
const int debug_net = 0;
int counter = 0;


enum
{
	HOSTAPD_MODE_IEEE80211B,
	HOSTAPD_MODE_IEEE80211G,
	HOSTAPD_MODE_IEEE80211A,
	NUM_HOSTAPD_MODES
};

void handle_receive();
void send_beacon();





void print_ip(unsigned char *data)
{
	print_number_nospace(data[0]);
	print_text(".");
	print_number_nospace(data[1]);
	print_text(".");
	print_number_nospace(data[2]);
	print_text(".");
	print_number(data[3]);
}

void print_mac(unsigned char *data)
{
	int i;
	print_hex2(data[0]);
	print_text(":");
	print_hex2(data[1]);
	print_text(":");
	print_hex2(data[2]);
	print_text(":");
	print_hex2(data[3]);
	print_text(":");
	print_hex2(data[4]);
	print_text(":");
	print_hex2(data[5]);
	print_text(" ");
}

void print_bss_conf(struct ieee80211_bss_conf *bss_conf)
{
	print_text("assoc=");
	print_number(bss_conf->assoc);
	print_text("ibss_joined=");
	print_number(bss_conf->ibss_joined);
	print_text("aid=");
	print_number(bss_conf->aid);
	print_text("use_cts_prot=");
	print_number(bss_conf->use_cts_prot);
	print_text("use_short_preamble=");
	print_number(bss_conf->use_short_preamble);
	print_text("use_short_slot=");
	print_number(bss_conf->use_short_slot);
	print_text("enable_beacon=");
	print_number(bss_conf->enable_beacon);
	print_text("dtim_period=");
	print_number(bss_conf->dtim_period);
	print_text("beacon_int=");
	print_number(bss_conf->beacon_int);
	print_text("assoc_capability=0x");
	print_hex(bss_conf->assoc_capability);
	print_text("basic_rates=0x");
	print_hex(bss_conf->basic_rates);
	print_text("ht_operation_mode=0x");
	print_hex(bss_conf->ht_operation_mode);
	print_text("channel_type=0x");
	print_hex(bss_conf->channel_type);
	print_text("arp_addr_cnt=0x");
	print_hex(bss_conf->arp_addr_cnt);
	print_text("arp_filter_enabled=0x");
	print_hex(bss_conf->arp_filter_enabled);
	print_text("qos=0x");
	print_hex(bss_conf->qos);
	print_text("idle=0x");
	print_hex(bss_conf->idle);
	print_text("ssid=");
	send_uart(bss_conf->ssid, bss_conf->ssid_len);
	print_lf();
	flush_uart();
}

unsigned char* get_mac()
{
	return net.hw.wiphy->perm_addr;
}

struct ieee80211_hw *ieee80211_alloc_hw(size_t priv_data_len,
					const struct ieee80211_ops *ops)
{
	net.hw.priv = kmalloc(priv_data_len, 0);
	net.hw.conf.channel = kmalloc(sizeof(struct ieee80211_channel), 0);
	net.hw.wiphy = kmalloc(sizeof(struct wiphy), 0);
	
	return &net.hw;
}

int hostapd_set_freq(int mode, 
	int freq,
	int channel, 
	int ht_enabled, 
	int sec_channel_offset,
	uint32_t changed)
{
	net.hw.conf.long_frame_max_tx_count = 4;
	struct ieee80211_channel *channel_ptr = net.hw.conf.channel;
	channel_ptr->hw_value = channel;
	
	rtl_ops.config(&net.hw, changed);
}




static uint8_t ieee802_11_erp_info(hapd)
{
	return 0x4;
}

void net_start_receive()
{
#ifdef ALWAYS_RX
	if(!USB_OTG_Core.rx_requested)
	{
		rtl_usb_receive(&net.hw, 
			0,
			0);
		USB_OTG_Core.rx_requested = 1;
	}
#else

	send_beacon();
// 	rtl_usb_receive();

#endif


//	net.current_function = handle_receive;
}

void net_start_transmit(unsigned char *data, 
	int size, 
	enum rtl_txq qnum,
	int need_resends)
{
/*
 * TRACE2
 * print_hex(data);
 * print_number(size);
 * flush_uart();
 */
	rtl_usb_tx(data,
		size,
		qnum);

	net.current_data = data;
	net.current_size = size;
	net.current_qnum = qnum;
	net.resends = need_resends;



	net.current_function = net_wait_tx;
	net.next_function = net_start_receive;
}

// Wait for a transmit to finish, then receive
void net_wait_tx()
{

	int ep_num = ep_mapping[net.current_qnum];
	int hc_num = net.hc_num_out[ep_num];
	if(HCD_GetURB_State(&USB_OTG_Core, hc_num) != URB_IDLE)
	{
//TRACE2
		if(net.resends)
		{
			rtl_usb_tx(net.current_data,
				net.current_size,
				net.current_qnum);
			net.resends--;
		}
		else
		{
//TRACE2
			net.current_function = net.next_function;
		}
	}


	
}



void i802_sta_deauth(const u8 *own_addr, 
	const u8 *addr,
	int reason)
{
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt*)net.out_buffer;
//TRACE

	bzero(mgmt, sizeof(struct ieee80211_mgmt));
	mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
					  WLAN_FC_STYPE_DEAUTH);
	memcpy(mgmt->da, addr, ETH_ALEN);
	memcpy(mgmt->sa, own_addr, ETH_ALEN);
	memcpy(mgmt->bssid, own_addr, ETH_ALEN);
	mgmt->u.deauth.reason_code = host_to_le16(reason);

TRACE
//print_buffer((u8*)mgmt, IEEE80211_HDRLEN + sizeof(mgmt->u.deauth) - 6);

	net_start_transmit((u8*)mgmt, 
		IEEE80211_HDRLEN + sizeof(mgmt->u.deauth) - 6, 
		RTL_TXQ_MGT,
		1);
// Called from net_init2 so must wait
	int ep_num = ep_mapping[RTL_TXQ_MGT];

TRACE
	while(HCD_GetURB_State(&USB_OTG_Core , net.hc_num_out[ep_num]) == URB_IDLE)
	{
		handle_usb();
		handle_uart();
	}
TRACE
}

void hostapd_sta_deauth(const u8 *addr,
	int reason)
{
	i802_sta_deauth(get_mac(), addr, reason);
}


static int hostapd_setup_bss()
{
	u8 addr[ETH_ALEN];
	memset(addr, 0xff, ETH_ALEN);
	hostapd_sta_deauth(addr, WLAN_REASON_PREV_AUTH_NOT_VALID);



	return 0;
}



void handle_auth(unsigned char *buf)
{
// send_auth_reply
	struct ieee80211_mgmt *reply = (struct ieee80211_mgmt *)net.out_buffer;
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)buf;

	if(debug_net)
	{
		TRACE
		print_text("authed ");
		print_mac(mgmt->sa);
	}

// Not directed at Marcy 2
	if(memcmp(net.hw.wiphy->perm_addr, mgmt->da, ETH_ALEN))
	{
		net_start_receive();
		return;
	}

	memcpy(net.authed_mac, mgmt->sa, ETH_ALEN);
	net.authed = 1;
	

	reply->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
	    WLAN_FC_STYPE_AUTH);
	memcpy(reply->da, mgmt->sa, ETH_ALEN);
	memcpy(reply->sa, net.hw.wiphy->perm_addr, ETH_ALEN);
	memcpy(reply->bssid, mgmt->bssid, ETH_ALEN);

	reply->u.auth.auth_alg = WLAN_AUTH_OPEN;
	reply->u.auth.auth_transaction = mgmt->u.auth.auth_transaction + 1;
	reply->u.auth.status_code = WLAN_STATUS_SUCCESS;

	net_start_transmit(net.out_buffer,
		IEEE80211_HDRLEN + sizeof(reply->u.auth),
		RTL_TXQ_MGT,
		1);
}

void handle_assoc(unsigned char *buf)
{
// send_assoc_resp
	struct ieee80211_mgmt *reply = (struct ieee80211_mgmt *)net.out_buffer;
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)buf;
	int reassoc = 0;

	if(!net.authed)
	{
		net_start_receive();
		return;
	}

// Not directed at Marcy 2
	if(memcmp(net.hw.wiphy->perm_addr, mgmt->da, ETH_ALEN))
	{
		net_start_receive();
		return;
	}

// Not authed
	if(memcmp(net.authed_mac, mgmt->sa, ETH_ALEN))
	{
		net_start_receive();
		return;
	}

	if(debug_net)
	{
		TRACE2
		print_text("associated with ");
		print_mac(mgmt->sa);
	}

	reply->frame_control =
		IEEE80211_FC(WLAN_FC_TYPE_MGMT,
			     (reassoc ? WLAN_FC_STYPE_REASSOC_RESP :
			      WLAN_FC_STYPE_ASSOC_RESP));
	memcpy(reply->da, mgmt->sa, ETH_ALEN);
	memcpy(reply->sa, net.hw.wiphy->perm_addr, ETH_ALEN);
	memcpy(reply->bssid, mgmt->bssid, ETH_ALEN);

//	reply->seq_ctrl = 0x30;
	reply->u.assoc_resp.capab_info = 0x0401;
	reply->u.assoc_resp.status_code = WLAN_STATUS_SUCCESS;
// TODO: Have to increment this for each client
// search for existing station
	int i;
	int aid = -1;
	for(i = 0; i < net.total_stations; i++)
	{
		if(!memcmp(net.station_macs[i], mgmt->sa, ETH_ALEN))
		{
			aid = i + 1;
			break;
		}
	}
	
	if(aid < 0)
	{
		if(net.total_stations < MAX_STATIONS)
		{
			net.station_macs[net.total_stations] = kmalloc(ETH_ALEN, 0);
			memcpy(net.station_macs[net.total_stations], mgmt->sa, ETH_ALEN);
			net.total_stations++;
			aid = net.total_stations;
		}
		else
		{
			TRACE2
			print_text("MAX_STATIONS reached.");
			net_start_receive();
			return;
		}
	}
	
	reply->u.assoc_resp.aid = 0xc000 + aid;

// Stuff
	const unsigned char stuff[] =
	{
		0x01, 0x08, 0x82, 0x84, 0x8b, 0x96, 0x0c, 0x12, 
		0x18, 0x24, 0x32, 0x04, 0x30, 0x48, 0x60, 0x6c
	};
	memcpy(reply->u.assoc_resp.variable, stuff, sizeof(stuff));
	


	net_start_transmit((unsigned char*)reply,
		IEEE80211_HDRLEN + sizeof(reply->u.assoc_resp) + sizeof(stuff),
		RTL_TXQ_MGT,
		1);
	net.connected = 1;
}


// for testing
void handle_80211_beacon(unsigned char *buf, int len)
{
	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)buf;
	unsigned char *beacon = mgmt->u.beacon.variable;

//	TRACE2
//	print_number((unsigned char*)&mgmt->u.beacon.variable - buf);
//	flush_uart();
//	print_buffer(buf, len);
//	flush_uart();
	
//	net_start_receive();
	send_beacon();
}

void send_beacon()
{
	int current_time = copter.timer_high;
	int difference = current_time - net.last_beacon;

// too soon after the last beacon
	if(difference >= 0 && difference < MAIN_TIMER_HZ / 10)
	{
//TRACE2
 		rtl_usb_receive();
		net.current_function = handle_receive;
		return;
	}
//TRACE2

	net.last_beacon = current_time;


	struct ieee80211_mgmt *mgmt = (struct ieee80211_mgmt *)net.out_buffer;
	unsigned char *pos = mgmt->u.beacon.variable;
	
	bzero(mgmt, ARM_OUT_BUFFER_SIZE);
	mgmt->frame_control = 
		IEEE80211_FC(WLAN_FC_TYPE_MGMT, WLAN_FC_STYPE_BEACON);
	memset(mgmt->da, 0xff, ETH_ALEN);
	memcpy(mgmt->sa, get_mac(), ETH_ALEN);
	memcpy(mgmt->bssid, get_mac(), ETH_ALEN);

	mgmt->u.beacon.timestamp = 0x5a2a3d2157LL;
	mgmt->u.beacon.beacon_int = 100;
	mgmt->u.beacon.capab_info = 0x0401;
	
	*pos++ = WLAN_EID_SSID;
	*pos++ = copter.ssid_len;
	memcpy(pos, copter.ssid, copter.ssid_len);
	pos += copter.ssid_len;
	
// HACK copy of a hostapd beacon
	const unsigned char beacon_data[] = 
	{ 

// WLAN_EID_SUPP_RATES
		0x01, 0x08, 0x82, 0x84, 0x8b, 0x96, 0x0c, 0x12, 0x18, 0x24, 

// WLAN_EID_DS_PARAMS
		0x03, 0x01, 0x08, 

// WLAN_EID_TIM
		0x05, 0x04, 0x01, 0x02, 0x00, 0x00, 

// WLAN_EID_ERP_INFO
		0x2a, 0x01, 0x04,

// WLAN_EID_EXT_SUPP_RATES
		0x32, 0x04, 0x30, 0x48, 0x60, 0x6c,

	};

	memcpy(pos, beacon_data, sizeof(beacon_data));
	pos[12] = copter.channel;
	pos += sizeof(beacon_data);




	net_start_transmit(net.out_buffer,
		pos - net.out_buffer,
		RTL_TXQ_MGT,
		1);
}



void handle_probe_req(unsigned char *buf)
{
//TRACE2
// handle_probe_req
	struct ieee80211_mgmt *mgmt = 
		(struct ieee80211_mgmt *)buf;
	struct ieee80211_mgmt *resp = 
		(struct ieee80211_mgmt*)net.out_buffer;
	unsigned char *pos = 0;
	bzero(resp, ARM_OUT_BUFFER_SIZE);

	resp->frame_control = 
		IEEE80211_FC(WLAN_FC_TYPE_MGMT,
			WLAN_FC_STYPE_PROBE_RESP);
	memcpy(resp->da, mgmt->sa, ETH_ALEN);
	memcpy(resp->sa, get_mac(), ETH_ALEN);
	memcpy(resp->bssid, get_mac(), ETH_ALEN);

	resp->u.probe_resp.beacon_int = BEACON_INT;
	resp->u.probe_resp.capab_info = 0x401;

/*
 * TRACE
 * print_text("ssid_len=");
 * print_number(net.ssid_len);
 * print_text(net.ssid);
 * flush_uart();
 */

	pos = resp->u.probe_resp.variable;
	*pos++ = WLAN_EID_SSID;
	*pos++ = copter.ssid_len;
	memcpy(pos, copter.ssid, copter.ssid_len);
	pos += copter.ssid_len;



// more HACK
	const unsigned char more_data[] = 
	{
// WLAN_EID_SUPP_RATES
// field type, field len, data
		0x01, 0x08, 0x82, 0x84, 0x8b, 0x96, 0x0c, 0x12, 
		0x18, 0x24, 

// WLAN_EID_DS_PARAMS
		0x03, 0x01, 0xff, 

// WLAN_EID_ERP_INFO
		0x2a, 0x01, 0x04, 

// WLAN_EID_EXT_SUPP_RATES
		0x32, 0x04, 0x30, 0x48, 0x60, 0x6c
	};

	memcpy(pos, more_data, sizeof(more_data));
	pos[12] = copter.channel;
	pos += sizeof(more_data);

	net_start_transmit(net.out_buffer,
		pos - net.out_buffer,
		RTL_TXQ_MGT,
		1);
}


// From contiki
static u16 chksum(u16 sum, const u8 *data, u16 len)
{
  u16 t;
  const u8 *dataptr;
  const u8 *last_byte;

  dataptr = data;
  last_byte = data + len - 1;
  
  while(dataptr < last_byte) {	/* At least two more bytes */
    t = (dataptr[0] << 8) + dataptr[1];
    sum += t;
    if(sum < t) {
      sum++;		/* carry */
    }
    dataptr += 2;
  }
  
  if(dataptr == last_byte) {
    t = (dataptr[0] << 8) + 0;
    sum += t;
    if(sum < t) {
      sum++;		/* carry */
    }
  }

  /* Return sum in host byte order. */
  return sum;
}


// Adds headers to the memory before data & sends it.
// Data needs to have memory allocated before it & stay untouched until
// transmission is done.
void send_udp(int size,
	void (*next_function)(),
	int need_resends)
{
// Expand the camera buffer if the header expands
 	unsigned char *buffer = net.out_buffer - 60;
	struct ieee80211_hdr *dst_hdr = (struct ieee80211_hdr*)buffer;

	dst_hdr->frame_control = IEEE80211_FTYPE_DATA | 
		IEEE80211_STYPE_DATA |
		IEEE80211_FCTL_FROMDS;

//#ifdef COPTER_MODE
#if 1
// dst
	memcpy(dst_hdr->addr1, net.src_mac, ETH_ALEN);
// src
	memcpy(dst_hdr->addr2, get_mac(), ETH_ALEN);
// bss
//	memcpy(dst_hdr->addr3, get_mac(), ETH_ALEN);
	memcpy(dst_hdr->addr3, net.filter_mac, ETH_ALEN);
#else 

// GROUND_MODE
// HACK: send directly from the camera to the laptop by hard coding everything
// should use the incoming addresses.
const unsigned char copter_mac[] = { 0x00, 0x02, 0x72, 0xa9, 0x31, 0x2c };
const unsigned char laptop_mac[] = { 0x94, 0xdb, 0xc9, 0xb0, 0xa1, 0x1b };

// dst
	memcpy(dst_hdr->addr1, laptop_mac, ETH_ALEN);
// src
	memcpy(dst_hdr->addr2, copter_mac, ETH_ALEN);
// bss
	memcpy(dst_hdr->addr3, get_mac(), ETH_ALEN);


#endif






/*
 * TRACE2
 * print_mac(dst_hdr->addr1);
 * print_mac(dst_hdr->addr2);
 * print_mac(dst_hdr->addr3);
 */


	const unsigned char extra_data[] = 
	{
		0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00, 0x08, 0x00
	};

	unsigned char *extra_data_begin = buffer + 24;
	memcpy(extra_data_begin, extra_data, sizeof(extra_data));
// IP section begins
	unsigned char *ip_buf = extra_data_begin + sizeof(extra_data);
// version, header length
	ip_buf[0] = 0x45;
// Flags
	ip_buf[1] = 0x10;
// Length from version byte
	ip_buf[2] = 0x0;
	ip_buf[3] = 0x0;
// ID
	ip_buf[4] = 0x0;
	ip_buf[5] = 0x0;
// Fragment offset
	ip_buf[6] = 0x0;
	ip_buf[7] = 0x0;
// Time to live
	ip_buf[8] = 0x80;
// Protocol
	ip_buf[9] = 0x11;
// IP checksum
	ip_buf[10] = 0x0;
	ip_buf[11] = 0x0;
// Source IPV4 address
	memcpy(ip_buf + 12, net.self_ip, 4);
// Dest IPV4 address
	memcpy(ip_buf + 16, net.src_ip, 4);
	
	
// UDP section begins
	unsigned char *udp_buf = ip_buf + 20;
// source port, big endian
	udp_buf[0] = (net.dst_port >> 8) & 0xff;
	udp_buf[1] = net.dst_port & 0xff;
// Dest port
//TRACE
//print_number(net.src_port);
	udp_buf[2] = (net.src_port >> 8) & 0xff;
	udp_buf[3] = net.src_port & 0xff;
// Length, starting from source port
	udp_buf[4] = 0;
	udp_buf[5] = 0;
// optional UDP chksum
	udp_buf[6] = 0;
	udp_buf[7] = 0;




// Data
// udp_data is 60 bytes after start of buffer
	unsigned char *udp_data = udp_buf + 8;


#ifdef UNCOMPRESSED
// Pack Y pixels only
	memcpy(udp_data + FRAME_HEADER_SIZE, camera.frame_buffer + net.frame_offset, payload_size - 4);
	net.frame_offset += payload_size - 4;
	if(net.frame_offset > FRAME_W * FRAME_H)
		net.frame_offset = 0;
#endif









// Compute IP size
	int ip_size = udp_data + size - ip_buf;
	ip_buf[2] = (ip_size >> 8) & 0xff;
	ip_buf[3] = ip_size & 0xff;

// Compute UDP size
	int udp_size = 8 + size;
	udp_buf[4] = (udp_size >> 8) & 0xff;
	udp_buf[5] = udp_size & 0xff;

// Compute IP header chksum
	u16 sum = 0;
	sum = chksum(sum, ip_buf, 20);
	sum ^= 0xffff;
	if(sum == 0x0) sum = 0xffff;
	ip_buf[10] = (sum >> 8) & 0xff;
	ip_buf[11] = sum & 0xff;

	int packet_size = udp_buf + 8 + size - buffer;


/*
 * TRACE2
 * print_text("packet_size=");
 * print_number(packet_size);
 */


 	net_start_transmit(buffer,
 		packet_size,
 		RTL_TXQ_MGT,
		need_resends);
	net.next_function = next_function;
}


void handle_dhcp(unsigned char *udp_buf, int udp_len)
{
	if(debug_net)
	{
		TRACE2
		print_text("DHCP from ");
		print_mac(net.src_mac);
	}

//print_buffer(udp_buf, udp_len);
	unsigned char *data_in = udp_buf + 8;
	
	if(data_in[0] == 0x1 &&
		data_in[1] == 0x1 &&
		data_in[2] == 0x6 &&
		data_in[3] == 0x0)
	{
// Discover
		if(data_in[240] == 53)
		{
// send offer
			int offset = 0;
			memset(net.out_buffer, 0, ARM_OUT_BUFFER_SIZE);
// Message type: boot reply
			net.out_buffer[offset++] = 0x02;
// hardware type: ethernet
			net.out_buffer[offset++] = 0x01;
// hardware address length
			net.out_buffer[offset++] = 0x06;
// hops
			net.out_buffer[offset++] = 0x00;
// transaction ID from incoming packet
			memcpy(net.out_buffer + offset, data_in + offset, 4);
			offset += 4;

// seconds elapsed			
			net.out_buffer[offset++] = 0x0;
			net.out_buffer[offset++] = 0x0;
// flags
			net.out_buffer[offset++] = 0x0;
			net.out_buffer[offset++] = 0x0;

// Old client address
			memset(net.out_buffer + offset, 0, 4);
			offset += 4;
// new client address is the station index + 3
			memcpy(net.out_buffer + offset, net.self_ip, 3);
			net.out_buffer[offset + 3] = net.last_station + 3;
			offset += 4;
// next server address
			memset(net.out_buffer + offset, 0, 4);
			offset += 4;
// relay address
			memcpy(net.out_buffer + offset, net.self_ip, 4);
			offset += 4;
// client MAC address
			memcpy(net.out_buffer + offset, net.src_mac, ETH_ALEN);
			offset += 6;
// padding, server hostname, boot filename
			offset += 202;

// magic cookie
			net.out_buffer[offset++] = 0x63;
			net.out_buffer[offset++] = 0x82;
			net.out_buffer[offset++] = 0x53;
			net.out_buffer[offset++] = 0x63;

// DHCP ACK
			net.out_buffer[offset++] = 53;
			net.out_buffer[offset++] = 1;

// Got a request
			if(data_in[offset] == 3)
			{
// ACK
				net.out_buffer[offset++] = 5;
			}
			else
// Got a discover
// offer
			{
				net.out_buffer[offset++] = 2;
			}

// server identifier: Marcy 2 address
			net.out_buffer[offset++] = 54;
			net.out_buffer[offset++] = 4;
			memcpy(net.out_buffer + offset, net.self_ip, 4);
			offset += 4;
// Point the response packet to the right addresses
			net.src_ip = net.out_buffer + offset;
			net.dst_ip = net.self_ip;

// lease time
			net.out_buffer[offset++] = 51;
			net.out_buffer[offset++] = 4;
			net.out_buffer[offset++] = 0x00;
			net.out_buffer[offset++] = 0x01;
			net.out_buffer[offset++] = 0x51;
			net.out_buffer[offset++] = 0x80;
// netmask
			net.out_buffer[offset++] = 1;
			net.out_buffer[offset++] = 4;
			net.out_buffer[offset++] = 255;
			net.out_buffer[offset++] = 255;
			net.out_buffer[offset++] = 255;
			net.out_buffer[offset++] = 0;

// broadcast address
//			net.out_buffer[offset++] = 28;
//			net.out_buffer[offset++] = 4;
//			net.out_buffer[offset++] = net.self_ip[0];
//			net.out_buffer[offset++] = net.self_ip[1];
//			net.out_buffer[offset++] = net.self_ip[2];
//			net.out_buffer[offset++] = 255;

// router
			net.out_buffer[offset++] = 3;
			net.out_buffer[offset++] = 4;
			net.out_buffer[offset++] = net.self_ip[0];
			net.out_buffer[offset++] = net.self_ip[1];
			net.out_buffer[offset++] = net.self_ip[2];
			net.out_buffer[offset++] = net.self_ip[3];

// Domain name offset
			net.out_buffer[offset++] = 15;
			
// string length
			int len = strlen(domain_text);
			net.out_buffer[offset++] = len;
			memcpy(net.out_buffer + offset, domain_text, len);
			offset += len;

// DNS server
			net.out_buffer[offset++] = 6;
			net.out_buffer[offset++] = 4;
			net.out_buffer[offset++] = net.self_ip[0];
			net.out_buffer[offset++] = net.self_ip[1];
			net.out_buffer[offset++] = net.self_ip[2];
			net.out_buffer[offset++] = net.self_ip[3];

// MTU
//			net.out_buffer[offset++] = 26;
//			net.out_buffer[offset++] = 2;
//			net.out_buffer[offset++] = 0x02;
//			net.out_buffer[offset++] = 0x40;

// end options
			net.out_buffer[offset++] = 0xff;
			offset += 26;

//print_buffer(net.out_buffer, offset);
			send_udp(offset,
				net_start_receive,
				4);
			return;
		}
		else
		{
			TRACE2
			print_text("unknown DHCP request ");
			print_number(data_in[240]);
		}
	}

	net_start_receive();
}

// TODO: certain sizes of domain name crash USB
// archive.ubuntu.com crashes it
// For now, it just aborts if the domain name doesn't match.
void handle_dns(unsigned char *ip_buf, unsigned char *udp_buf, int udp_len)
{
	int offset = 0;
	int i = 0;
	const int debug = debug_net;

// Dst address is Marcy2
	if(!memcmp(net.dst_ip, net.self_ip, 4))
	{
//TRACE2
//print_text("DNS ");
//print_buffer(udp_buf, udp_len);
		unsigned char *dns_buf = udp_buf + 8;
		
		int flags = (dns_buf[2] << 8) |
			dns_buf[3];
		int got_it = 0;
		int type = 0;
// Standard query
		if(flags == 0x0100)
		{
			offset = 12;
			

// Skip to end of name to get type of record
			for(i = offset; dns_buf[i] != 0; i++)
			{
				i += dns_buf[i];
			}

			i++;
			type = (dns_buf[i] << 8) | dns_buf[i + 1];

			
// Type A == marcy2
// Don't compare final 0x0
			if(type == 0x1 && !memcmp(dns_buf + offset, (char*)net.a_record, net.a_record_len - 1) ||
				type == 0xc && !memcmp(dns_buf + offset, (char*)net.ptr_record, net.ptr_record_len - 1))
			{
				got_it = 1;
			}

			if(debug_net)
			{
				TRACE2
				print_text("type=0x");
				print_hex(type);
				print_text("got_it=");
				print_hex(got_it);
			}

			if(got_it)
			{
				flags = 0x8180;
			}
			else
			{
// Send not found
				flags = 0x8183;
			}




			offset = 0;
			unsigned char *reply = net.out_buffer;
// transaction ID
			reply[offset] = dns_buf[offset];
			offset++;
			reply[offset] = dns_buf[offset];
			offset++;
// flags
			reply[offset++] = ((flags >> 8) & 0xff);
			reply[offset++] = (flags & 0xff);
// questions
			reply[offset++] = 0x00;
			reply[offset++] = 0x01;
// answer RRs
			reply[offset++] = 0x00;
			reply[offset++] = got_it;
// authority RRs
			reply[offset++] = 0x00;
			reply[offset++] = 0x00;
// additional RRs
			reply[offset++] = 0x00;
			reply[offset++] = 0x00;


// original query
			int counter = 0;
			if(debug) print_text("query=");
			for(i = 12; dns_buf[i] != 0; i++)
			{
				if(debug) 
				{
					if(counter > 0)
					{
						send_uart(dns_buf + i, 1);
						counter--;
					}
					else
					{
						counter = dns_buf[i];
						if(i != 12) send_uart(".", 1);
					}
				}

				reply[offset++] = dns_buf[i];
			}
			if(debug) send_uart(" ", 1);



// end of string
			reply[offset++] = 0;
// type
			reply[offset++] = type >> 8;
			reply[offset++] = type & 0xff;
// class IN
			reply[offset++] = 0x00;
			reply[offset++] = 0x01;





			if(got_it)
			{
// authoritative nameserver
// 0xc0, offset of name after start of out_buffer
				reply[offset++] = 0xc0;
				reply[offset++] = 0x0c;
// type
				reply[offset++] = type >> 8;
				reply[offset++] = type & 0xff;
// class: IN
				reply[offset++] = 0x00;
				reply[offset++] = 0x01;
// TTL in seconds
				reply[offset++] = 0x00;
				reply[offset++] = 0x00;
				reply[offset++] = 0x0e;
				reply[offset++] = 0x10;

// A record
				if(type == 0x1)
				{
// length
					reply[offset++] = 0x00;
					reply[offset++] = 0x04;
// IP address
					memcpy(reply + offset, net.self_ip, 4);
					offset += 4;
				}
				else
// PTR
				if(type == 0xc)
				{
// length
					reply[offset++] = net.a_record_len >> 8;
					reply[offset++] = net.a_record_len & 0xff;
					
					memcpy(reply + offset, net.a_record, net.a_record_len);
					offset += net.a_record_len;
				}
			}

//TRACE2
//print_text("size=");
//print_number(offset);
//print_buffer(reply, offset);
// HACK to get USB to work
			if(offset % 64)
			{
				bzero(reply + offset, 64 - (offset % 64));
				offset += 64 - (offset % 64);
			}

			send_udp(offset,
				net_start_receive,
				1);
			return;
		}
		else
// Don't handle other queries
		{
			net_start_receive();
			return;
		}
	}

// Not our dst address
	net_start_receive();
}

void handle_udp(unsigned char *ip_buf, int ip_len)
{
	unsigned char *udp_buf = ip_buf + 20;
	net.total_udp++;

	net.src_ip = ip_buf + 12;
	net.dst_ip = ip_buf + 16;
	net.src_port = (udp_buf[0] << 8) | udp_buf[1];
	net.dst_port = (udp_buf[2] << 8) | udp_buf[3];

//TRACE
//print_number(net.dst_port);

	switch(net.dst_port)
	{
// DHCP port
		case 67:
			handle_dhcp(udp_buf, ip_len - 20);
			break;

// DNS port
		case 53:
			handle_dns(ip_buf, udp_buf, ip_len - 20);
			break;

// Beacon port
		case MARCY2_PORT:
// beacons only go to copter
			if(!memcmp(net.dst_ip, (unsigned char*)copter_ip, 4))
			{
				handle_beacon(udp_buf + 8);
			}
			else
			{
				net_start_receive();
			}
			break;

		default:
			net_start_receive();
			break;
	}
}

void send_tcp(int payload_size,
	unsigned int sequence_number,
	unsigned int ack_number,
	unsigned int flags,
	void (*next_function)())
{
//TRACE
// Adjust this so the headers line up with the existing data
	unsigned char *reply;
	int offset = 0;

// Rewind by TCP header, IP header, & 802.11 header
	if(flags == 0x12)
	{
		reply = net.out_buffer - 44 - 20 - 32;
	}
	else
	{
		reply = net.out_buffer - 32 - 20 - 32;
	}

	struct ieee80211_hdr *dst_hdr = (struct ieee80211_hdr*)reply;

	dst_hdr->frame_control = IEEE80211_FTYPE_DATA | 
		IEEE80211_STYPE_DATA |
		IEEE80211_FCTL_FROMDS;
// dst
	memcpy(dst_hdr->addr1, net.src_mac, ETH_ALEN);
// src
	memcpy(dst_hdr->addr2, get_mac(), ETH_ALEN);
// bss
	memcpy(dst_hdr->addr3, get_mac(), ETH_ALEN);

	const unsigned char extra_data[] = 
	{
		0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00, 0x08, 0x00
	};

	unsigned char *extra_data_begin = reply + 24;
	memcpy(extra_data_begin, extra_data, sizeof(extra_data));
	
// IP section begins
	unsigned char *ip_buf = extra_data_begin + sizeof(extra_data);
// version, header length
	ip_buf[0] = 0x45;
// Flags
	ip_buf[1] = 0x20;
// Length from version byte
	ip_buf[2] = 0x0;
	ip_buf[3] = 0x0;
// ID
	ip_buf[4] = 0x0;
	ip_buf[5] = 0x0;
// Fragment offset + flags
	ip_buf[6] = 0x40;
	ip_buf[7] = 0x00;
// Time to live
	ip_buf[8] = 0x80;
// Protocol
	ip_buf[9] = 0x06;
// IP checksum
	ip_buf[10] = 0x0;
	ip_buf[11] = 0x0;
// Source IPV4 address
	memcpy(ip_buf + 12, net.self_ip, 4);
// Dest IPV4 address
	memcpy(ip_buf + 16, net.src_ip, 4);

// TCP section begins
	unsigned char *tcp_buf = ip_buf + 20;
// source port, big endian
	tcp_buf[offset++] = (net.dst_port >> 8) & 0xff;
	tcp_buf[offset++] = net.dst_port & 0xff;
// Dest port
	tcp_buf[offset++] = (net.src_port >> 8) & 0xff;
	tcp_buf[offset++] = net.src_port & 0xff;

// sequence number
	tcp_buf[offset++] = (sequence_number >> 24) & 0xff;
	tcp_buf[offset++] = (sequence_number >> 16) & 0xff;
	tcp_buf[offset++] = (sequence_number >> 8) & 0xff;
	tcp_buf[offset++] = sequence_number & 0xff;

// ack number
	tcp_buf[offset++] = (ack_number >> 24) & 0xff;
	tcp_buf[offset++] = (ack_number >> 16) & 0xff;
	tcp_buf[offset++] = (ack_number >> 8) & 0xff;
	tcp_buf[offset++] = ack_number & 0xff;

// Header length
	if(flags == 0x12)
	{
		tcp_buf[offset++] = 0xb0;
	}
	else
	{
		tcp_buf[offset++] = 0x80;
	}
// flags
	tcp_buf[offset++] = flags;

// Window size / 32
	tcp_buf[offset++] = 0x08;
	tcp_buf[offset++] = 0x00;
	
// chksum
	int tcp_chksum_offset = offset;
	tcp_buf[offset++] = 0x00;
	tcp_buf[offset++] = 0x00;

// blank
	tcp_buf[offset++] = 0x00;
	tcp_buf[offset++] = 0x00;

// options
// If a SYN ACK, put in different options
	if(flags == 0x12)
	{
	// maximum segment size
		tcp_buf[offset++] = 0x02;
		tcp_buf[offset++] = 0x04;
		tcp_buf[offset++] = 0x05;
		tcp_buf[offset++] = 0xb4;
	// NOP
		tcp_buf[offset++] = 0x1;
	// window scale 5 -> 32
		tcp_buf[offset++] = 0x03;
		tcp_buf[offset++] = 0x03;
		tcp_buf[offset++] = 0x05;
	// NOP
		tcp_buf[offset++] = 0x1;
	// NOP
		tcp_buf[offset++] = 0x1;
	// Time stamp
		tcp_buf[offset++] = 0x08;
		tcp_buf[offset++] = 0x0a;

	// tsval
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;

	// tsecr
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;

	// SACK permitted
		tcp_buf[offset++] = 0x04;
		tcp_buf[offset++] = 0x02;

		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
	}
	else
	{
		tcp_buf[offset++] = 0x1;
		tcp_buf[offset++] = 0x1;
	// Time stamp
		tcp_buf[offset++] = 0x08;
		tcp_buf[offset++] = 0x0a;

	// tsval
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;

	// tsecr
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
		tcp_buf[offset++] = 0x00;
	}

/*
 * TRACE
 * print_text("flags=0x");
 * print_hex(flags);
 * print_text("offset=");
 * print_number(offset);
 */

// IP length from version byte
	int length = offset + 20 + payload_size;
	ip_buf[2] = (length >> 8) & 0xff;
	ip_buf[3] = length & 0xff;


// compute TCP chksum
	u16 sum = 0;
// TCP segment length + protocol code with no carry
	sum = offset + payload_size + 0x6;
// src & dst IP addresses
	sum = chksum(sum, ip_buf + 12, 8);
// TCP header & data
	sum = chksum(sum, tcp_buf, offset + payload_size);
	sum ^= 0xffff;
	if(sum == 0x0) sum = 0xffff;
	tcp_buf[tcp_chksum_offset++] = (sum >> 8) & 0xff;
	tcp_buf[tcp_chksum_offset++] = sum & 0xff;


	

// Compute IP header chksum
	sum = 0;
	sum = chksum(sum, ip_buf, 20);
	sum ^= 0xffff;
	if(sum == 0x0) sum = 0xffff;
	ip_buf[10] = (sum >> 8) & 0xff;
	ip_buf[11] = sum & 0xff;



	int packet_size = tcp_buf + offset + payload_size - reply;

// HACK to get USB to work
	if(packet_size % 64)
	{
		bzero(reply + packet_size, 64 - (packet_size % 64));
		packet_size += 64 - (packet_size % 64);
	}

 	net_start_transmit(reply,
 		packet_size,
 		RTL_TXQ_MGT,
		1);


/*
 * TRACE2
 * print_text("packet_size=");
 * print_number(packet_size);
 */



	net.next_function = next_function;
}


//#ifdef COPTER_MODE
void handle_tcp2()
{
//TRACE2
// web server request
	if(net.dst_port == 80)
	{
//send_uart(net.tcp_payload, net.tcp_size);
		
		handle_http(net.tcp_payload, net.tcp_size);
		
		
	}
	else
	{
		net_start_receive();
	}
	return;
}

void handle_tcp(unsigned char *ip_buf, int ip_len)
{
	net.total_tcp++;
	net.src_ip = ip_buf + 12;
	net.dst_ip = ip_buf + 16;

// Get accurate IP len from the header
	ip_len = (ip_buf[2] << 8) |
		ip_buf[3];

// Wrong destination IP address
	if(memcmp(net.dst_ip, net.self_ip, 4))
	{
		net_start_receive();
		return;
	}


	unsigned char *tcp_buf = ip_buf + 20;
	int tcp_len = ip_len - 20;

	net.src_port = (tcp_buf[0] << 8) | tcp_buf[1];
	net.dst_port = (tcp_buf[2] << 8) | tcp_buf[3];

	int flags = tcp_buf[13];
	net.ack_number = (tcp_buf[4] << 24) |
		(tcp_buf[5] << 16) |
		(tcp_buf[6] << 8) |
		tcp_buf[7];
	net.sequence_number = (tcp_buf[8] << 24) |
		(tcp_buf[9] << 16) |
		(tcp_buf[10] << 8) |
		tcp_buf[11];

TRACE2
print_text("from ");
print_ip(net.src_ip);
print_text("port=");
print_number(net.dst_port);
print_text("type=");


// SYN
	if(flags == 0x02)
	{
// send SYN ACK & create new connection
// TODO: sequence number needs to be random
		net.sequence_number = 0x12345678;
		net.ack_number++;
print_text("SYN\n");
		
		send_tcp(0,
			net.sequence_number,
			net.ack_number,
			0x12,
			net_start_receive);
	}
	else
// FIN, ACK
	if(flags == 0x11)
	{
		net.ack_number++;
print_text("FIN\n");
// Send ACK
		send_tcp(0,
			net.sequence_number,
			net.ack_number,
			0x10,
			net_start_receive);
	}
	else
// PSH, ACK
	if(flags == 0x18)
	{
print_text("PSH\n");
// Store payload for handler
		net.tcp_payload = tcp_buf + 32;
		net.tcp_size = tcp_len - 32;
// Increment ACK number by payload size
		net.ack_number += net.tcp_size;

// Send ACK
		send_tcp(0,
			net.sequence_number,
			net.ack_number,
			0x10,
			handle_tcp2);
	}
	else
// Ignore ACK
	{

if(flags == 0x10) print_text("ACK\n");
		net_start_receive();
		return;
	}
}


void handle_icmp(unsigned char *ip_buf, int ip_len)
{
	net.total_icmp++;
	net.src_ip = ip_buf + 12;
	net.dst_ip = ip_buf + 16;

// not directed at Marcy 2
	if(memcmp(net.dst_ip, net.self_ip, 4))
	{
		net_start_receive();
		return;
	}

	if(debug_net)
	{
		TRACE2
		print_text("from ");
		print_ip(net.src_ip);
	}

// 802.11 header
 	unsigned char *buffer = net.out_buffer;
	struct ieee80211_hdr *dst_hdr = (struct ieee80211_hdr*)buffer;

	dst_hdr->frame_control = IEEE80211_FTYPE_DATA | 
		IEEE80211_STYPE_DATA |
		IEEE80211_FCTL_FROMDS;
// dst
	memcpy(dst_hdr->addr1, net.src_mac, ETH_ALEN);
// src
	memcpy(dst_hdr->addr2, net.dst_mac, ETH_ALEN);
// bss
	memcpy(dst_hdr->addr3, net.filter_mac, ETH_ALEN);

/*
 * TRACE2
 * print_mac(dst_hdr->addr1);
 * print_mac(dst_hdr->addr2);
 * print_mac(dst_hdr->addr3);
 */

	const unsigned char extra_data[] = 
	{
		0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00, 
// Type: IP
		0x08, 0x00
	};

	unsigned char *extra_data_begin = buffer + 24;
	memcpy(extra_data_begin, extra_data, sizeof(extra_data));


	unsigned char *ip_buf_out = buffer + 32;
// version
	ip_buf_out[0] = ip_buf[0];
// type
	ip_buf_out[1] = ip_buf[1];
// length
	ip_buf_out[2] = ip_buf[2];
	ip_buf_out[3] = ip_buf[3];
// ID
	ip_buf_out[4] = ip_buf[4];
	ip_buf_out[5] = ip_buf[5];
// flags & offset
	ip_buf_out[6] = ip_buf[6];
	ip_buf_out[7] = ip_buf[7];
// TTL
	ip_buf_out[8] = ip_buf[8];
// protocol
	ip_buf_out[9] = ip_buf[9];
// chksum
	ip_buf_out[10] = 0;
	ip_buf_out[11] = 0;
// source IP address
	memcpy(ip_buf_out + 12, net.dst_ip, 4);
// dest IP address
	memcpy(ip_buf_out + 16, net.src_ip, 4);
	
// ICMP payload
// type 0 for reply
	ip_buf_out[20] = 0;
//	ip_buf_out[20] = 0x8;
// code
	ip_buf_out[21] = 0;
// chksum is identical to the old one with a different type field
	ip_buf_out[22] = ip_buf[22];
	ip_buf_out[23] = ip_buf[23];
// ID
	ip_buf_out[24] = ip_buf[24];
	ip_buf_out[25] = ip_buf[25];
// sequence number
	ip_buf_out[26] = ip_buf[26];
	ip_buf_out[27] = ip_buf[27];
// data
	memcpy(ip_buf_out + 28, ip_buf + 28, ip_len - 28);


// Compute IP header chksum
	u16 sum = 0;
	sum = chksum(sum, ip_buf_out, 20);
	sum ^= 0xffff;
	if(sum == 0x0) sum = 0xffff;
	ip_buf_out[10] = (sum >> 8) & 0xff;
	ip_buf_out[11] = sum & 0xff;

// Compute ICMP chksum
// chksum is identical to the old one with a different type field
	sum = (ip_buf[22] << 8) | ip_buf[23];
	if(sum > 0xffff - 0x0800)
	{
		sum += 0x0801;
	}
	else
	{
		sum += 0x0800;
	}

	ip_buf_out[22] = (sum >> 8) & 0xff;
	ip_buf_out[23] = sum & 0xff;

//TRACE
//print_buffer(buffer, ip_len + 32);

	net_start_transmit(buffer,
 		ip_len + 32,
 		RTL_TXQ_MGT,
		0);
}

void handle_arp(unsigned char *ip_buf, int ip_len)
{
// got request opcode
	if(ip_buf[6] == 0x00 &&
		ip_buf[7] == 0x01)
	{
		net.src_ip = ip_buf + 14;
		net.dst_ip = ip_buf + 24;
		net.total_arp++;

// Got Marcy 2 address
// Send reply
		if(!memcmp(net.dst_ip, net.self_ip, 4) ||
			!memcmp(net.dst_ip, (unsigned char*)ground_ip, 4))
		{

			if(debug_net)
			{
				TRACE2
				print_ip(net.src_ip);
				print_text("wants ");
				print_ip(net.dst_ip);
			}

// 802.11 header
 			unsigned char *buffer = net.out_buffer;
			struct ieee80211_hdr *dst_hdr = (struct ieee80211_hdr*)buffer;

			dst_hdr->frame_control = IEEE80211_FTYPE_DATA | 
				IEEE80211_STYPE_DATA |
				IEEE80211_FCTL_FROMDS;
// dst
			memcpy(dst_hdr->addr1, net.filter_mac, ETH_ALEN);
// src
			memcpy(dst_hdr->addr2, net.dst_mac, ETH_ALEN);
// bss
			memcpy(dst_hdr->addr3, net.filter_mac, ETH_ALEN);

			const unsigned char extra_data[] = 
			{
				0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00, 
// Type: ARP
				0x08, 0x06
			};

			unsigned char *extra_data_begin = buffer + 24;
			memcpy(extra_data_begin, extra_data, sizeof(extra_data));

			unsigned char *ip_buf_out = buffer + 32;
// HW type: ethernet
			ip_buf_out[0] = 0x00;
			ip_buf_out[1] = 0x01;
// protocol type: IP
			ip_buf_out[2] = 0x08;
			ip_buf_out[3] = 0x00;
// hardware size
			ip_buf_out[4] = 0x06;
// protocol size
			ip_buf_out[5] = 0x04;
// Opcode: reply
			ip_buf_out[6] = 0x00;
			ip_buf_out[7] = 0x02;

			if(!memcmp(net.dst_ip, net.self_ip, 4))
			{
	// Sender MAC address
				memcpy(ip_buf_out + 8, get_mac(), ETH_ALEN);
	// Sender IP address
				memcpy(ip_buf_out + 14, net.self_ip, 4);
			}
			else
			{

	// Sender MAC address
				memcpy(ip_buf_out + 8, net.ground_mac, ETH_ALEN);
	// Sender IP address
				memcpy(ip_buf_out + 14, ground_ip, 4);
			}


// Target MAC address
			memcpy(ip_buf_out + 18, net.src_mac, ETH_ALEN);
// Target IP address
			memcpy(ip_buf_out + 24, net.src_ip, 4);
// Trailer
			memset(ip_buf_out + 28, 0x0, 64 - 28);

			net_start_transmit(buffer,
				64,
// 				28 + 18 + 32,
 				RTL_TXQ_MGT,
				4);
		}
		else
		{
			net_start_receive();
		}
	}
	else
	{
		net_start_receive();
	}
}
//#endif // COPTER_MODE


#if 0
void forward_packet(unsigned char *buf, int len)
{
	memcpy(net.out_buffer, buf, len);
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr*)net.out_buffer;
	
// ground
	memcpy(hdr->addr1, net.filter_mac, ETH_ALEN);
	memcpy(hdr->addr2, get_mac(), ETH_ALEN);
// laptop
	memcpy(hdr->addr3, net.src_mac, ETH_ALEN);

/*
 * TRACE2
 * print_mac(hdr->addr1);
 * print_mac(hdr->addr2);
 * print_mac(hdr->addr3);
 */

	net_start_transmit(net.out_buffer,
 		len,
 		RTL_TXQ_MGT,
		0);
}
#endif // 0




void handle_data(unsigned char *buf, int len)
{
	struct ieee80211_hdr *hdr = (struct ieee80211_hdr*)buf;
	unsigned char *ip_buf = buf + 32;
	int ip_len = len - 32;
	net.total_data++;




	if(
//#ifdef COPTER_MODE
		!net.connected ||
//#endif
		ip_len <= 0)
	{
		net_start_receive();
		return;
	}

	net.dst_mac = hdr->addr1;
	net.src_mac = hdr->addr2;
	net.filter_mac = hdr->addr3;

	int i;





//TRACE2
//print_number(len);
//print_mac(hdr->addr1);
//print_mac(hdr->addr2);
//print_mac(hdr->addr3);







//#ifdef COPTER_MODE

// HACK: forward packets to & from the ground station
// 	if(!memcmp((unsigned char*)net.ground_mac, net.filter_mac, ETH_ALEN) ||
// 		!memcmp((unsigned char*)net.ground_mac, net.src_mac, ETH_ALEN)) 
// 	{
// 		forward_packet(buf, len);
// 		return;
// 	}

	int got_it = 0;
	for(i = 0; i < net.total_stations; i++)
	{
		if(!memcmp(net.station_macs[i], net.src_mac, ETH_ALEN)) 
		{
			got_it = 1;
			net.last_station = i;
			break;
		}
	}



// not assoced
	if(!got_it)
	{
		net_start_receive();
		return;
	}


// ARP
	if(buf[30] == 0x08 &&
		buf[31] == 0x06)
	{
		handle_arp(ip_buf, ip_len);
		return;
	}
//#endif // COPTER_MODE

// Destination not us
	if(memcmp(net.dst_mac, get_mac(), ETH_ALEN))
	{
		net_start_receive();
		return;
	}


/*
 * TRACE
 * print_buffer(ip_buf, ip_len);
 */

// IP protocol
	int protocol = ip_buf[9];
	switch(protocol)
	{
// UDP
		case 0x11:
			handle_udp(ip_buf, ip_len);
			break;


//#ifdef COPTER_MODE
// ICMP
		case 1:
			handle_icmp(ip_buf, ip_len);
			break;

// TCP
		case 0x06:
			handle_tcp(ip_buf, ip_len);
			break;
//#endif // COPTER_MODE



		default:
			net_start_receive();
			break;
	}

}

void handle_packet(unsigned char *data, int len)
{
	unsigned char *buf = data + PACKET_OFFSET;



	struct ieee80211_hdr *hdr = (struct ieee80211_hdr*)buf;
	u16 fc = hdr->frame_control;
	int type = WLAN_FC_GET_TYPE(fc);



/*
 * TRACE
 * print_text("fc=0x");
 * print_hex(fc);
 * print_text("WLAN_FC_GET_TYPE=");
 * print_number(type);
 */








	switch(type)
	{
		case IEEE80211_FTYPE_MGMT:
		{
// ieee802_11_mgmt
			u16 stype = WLAN_FC_GET_STYPE(fc) >> 4;
			net.total_mgmt++;

//TRACE2
//print_text("stype=");
//print_number(stype);



			switch(stype)
			{
// the copter is the access point
//#ifdef COPTER_MODE
				case WLAN_FC_STYPE_PROBE_REQ:
					handle_probe_req(buf);
					break;
				
				case WLAN_FC_STYPE_AUTH:
					handle_auth(buf);
					break;
				
//				case WLAN_FC_STYPE_DEAUTH:
//					handle_deauth(buf);
//					break;
				
//				case WLAN_FC_STYPE_BEACON:
//					handle_80211_beacon(buf, len);
//					break;
				
				case WLAN_FC_STYPE_ASSOC_REQ:
					handle_assoc(buf);
					break;
//#endif // COPTER_MODE

				default:
					net_start_receive();
					break;
			}
			break;
		}

//		case IEEE80211_FTYPE_CTL:
//			break;

		case IEEE80211_FTYPE_DATA:
			handle_data(buf, len - PACKET_OFFSET);
			break;
		
		default:
			net_start_receive();
			break;
	}
}





void handle_receive()
{
// Got an incoming packet
//char buffer[1];
//buffer[0] = USB_OTG_Core.user_buffer + 'A';
//send_uart(buffer, 1);

#ifndef ALWAYS_RX
	if(USB_OTG_Core.host.URB_State[net.hc_num_in] != URB_IDLE)
#else
	if(USB_OTG_Core.buffer_filled[USB_OTG_Core.user_buffer])
#endif
	{
		net.bytes_received = USB_OTG_Core.host.XferCnt[net.hc_num_in];
		if(USB_OTG_Core.host.URB_State[net.hc_num_in] == URB_ERROR)
			net.bytes_received = IN_BUFFER_SIZE;
		
// 		TRACE2
// 		print_text("URB_State=");
// 		print_number(USB_OTG_Core.host.URB_State[net.hc_num_in]);
// 		print_text("len=");
// 		print_number(bytes_received);
// 		print_text("counter=");
// 		print_number(counter);
//		counter = 0;
//		print_buffer(net.in_buffer, USB_OTG_Core.host.XferCnt[net.hc_num_in]);
//		print_buffer(net.in_buffer, 16);




#ifndef ALWAYS_RX
		if(net.bytes_received)
#else
		if(USB_OTG_Core.buffer_filled[USB_OTG_Core.user_buffer])
#endif
		{
#ifndef ALWAYS_RX


			handle_packet(net.in_buffer, net.bytes_received);
			net.total_rx += net.bytes_received;


#else

			int bytes_received = USB_OTG_Core.buffer_size[USB_OTG_Core.user_buffer];
			handle_packet(USB_OTG_Core.rx_buffers[USB_OTG_Core.user_buffer], 
				bytes_received);

// Release next buffer to interrupt handler
			USB_OTG_Core.buffer_filled[USB_OTG_Core.user_buffer] = 0;
			USB_OTG_Core.user_buffer = !USB_OTG_Core.user_buffer;
#endif
		}
#ifndef ALWAYS_RX
		else
		{
			net_start_receive();
		}
#endif
	}
	else
	{
		counter++;
	}
	
}

static void net_idle()
{
}


void init_net2()
{

	if(!usb.usb_initialized) return;

//TRACE
    net.hc_num_in = USBH_Alloc_Channel(&USB_OTG_Core, 
        USB_Host.device_prop.Ep_Desc[0][0].bEndpointAddress);
/* Open channel for IN endpoint */
// Corresponds to address 0x81
    USBH_Open_Channel(&USB_OTG_Core,
                        net.hc_num_in,
                        USB_Host.device_prop.address,
                        USB_Host.device_prop.speed,
                        EP_TYPE_BULK,
                        USB_Host.device_prop.Ep_Desc[0][0].wMaxPacketSize); 


// Corresponds to address 0x02
    net.hc_num_out[0] = USBH_Alloc_Channel(&USB_OTG_Core, 
        USB_Host.device_prop.Ep_Desc[0][1].bEndpointAddress);
/* Open channel for OUT endpoints */
    USBH_Open_Channel(&USB_OTG_Core,
                        net.hc_num_out[0],
                        USB_Host.device_prop.address,
                        USB_Host.device_prop.speed,
                        EP_TYPE_BULK,
                        USB_Host.device_prop.Ep_Desc[0][1].wMaxPacketSize); 

// Corresponds to address 0x03
    net.hc_num_out[1] = USBH_Alloc_Channel(&USB_OTG_Core, 
        USB_Host.device_prop.Ep_Desc[0][2].bEndpointAddress);
/* Open channel for OUT endpoints */
    USBH_Open_Channel(&USB_OTG_Core,
                        net.hc_num_out[1],
                        USB_Host.device_prop.address,
                        USB_Host.device_prop.speed,
                        EP_TYPE_BULK,
                        USB_Host.device_prop.Ep_Desc[0][2].wMaxPacketSize); 




    net.hc_num_int = USBH_Alloc_Channel(&USB_OTG_Core, 
        USB_Host.device_prop.Ep_Desc[0][3].bEndpointAddress);
/* Open channel for IN endpoint */
    USBH_Open_Channel(&USB_OTG_Core,
                        net.hc_num_int,
                        USB_Host.device_prop.address,
                        USB_Host.device_prop.speed,
                        EP_TYPE_INTR,
                        USB_Host.device_prop.Ep_Desc[0][3].wMaxPacketSize); 






	rtl_usb_probe();
//TRACE
// normally done in rtl_op_start
//TRACE
	rtl_usb_start(&net.hw);


	struct rtl_priv *priv = rtl_priv(&net.hw);
//TRACE
	struct rtl_mac *mac = &priv->mac80211;
//TRACE


	
	
	struct ieee80211_vif *vif = &net.vif;
//TRACE
	struct ieee80211_bss_conf *bss_conf = &vif->bss_conf;


//TRACE

	vif->type = NL80211_IFTYPE_AP;
	memcpy(vif->addr, net.hw.wiphy->perm_addr, ETH_ALEN);
//TRACE


//rtl92c_dm_watchdog(&net.hw);

	
	hostapd_set_freq(1, 2412, 1, 0, 0, 0xffffffff);
//TRACE

// Copied straight from a kernel trace, since we got tired of following hostapd
// -60dbm transmit power with it enabled
// -41dbm transmit power with it disabled
#if 0
	{
		struct ieee80211_tx_queue_params params = 
		{
			102, 7, 15, 2, 0
		};
		rtl_op_conf_tx(&net.hw, vif, 0, &params);
	}

	{
		struct ieee80211_tx_queue_params params = 
		{
			188, 15, 31, 2, 0
		};
		rtl_op_conf_tx(&net.hw, vif, 1, &params);
	}

	{
		struct ieee80211_tx_queue_params params = 
		{
			0, 31, 1023, 3, 0
		};
		rtl_op_conf_tx(&net.hw, vif, 2, &params);
	}

	{
		struct ieee80211_tx_queue_params params = 
		{
			0, 31, 1023, 7, 0
		};
		rtl_op_conf_tx(&net.hw, vif, 3, &params);
	}
#endif // 0

//TRACE
	rtl_op_add_interface(&net.hw, vif);
//TRACE

	bss_conf->use_short_slot = 0;
	bss_conf->enable_beacon = 0;
	bss_conf->dtim_period = 0;
	bss_conf->beacon_int = 0;
	bss_conf->basic_rates = 0x1;
	bss_conf->idle = 1;
	
	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0xe);
//TRACE



	bss_conf->basic_rates = 0xf;
	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x20);

//TRACE



// Change channel again
// don't think freq is used
	hostapd_set_freq(1, 2447, copter.channel, 0, 0, 0x40);

//TRACE



	hostapd_setup_bss();
//TRACE


	bss_conf->beacon_int = BEACON_INT;
	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x40);

//	bss_conf->enable_beacon = 0;
	bss_conf->enable_beacon = 1;
	bss_conf->dtim_period = 2;
	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x8300);

	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x2);
//TRACE

	bss_conf->use_short_slot = 1;
	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x8);
//TRACE

	rtl_op_bss_info_changed(&net.hw, vif, bss_conf, 0x4);

//TRACE

	net_start_receive();
//TRACE
	print_text("\nNetwork initialized.");
//TRACE

}

void show_net_status()
{
	TRACE2
	print_text("net.current_function=");
	print_hex(net.current_function);
	print_text("\ntotal_tx=");
	print_number(net.total_tx);
	print_text("\ntotal_rx=");
	print_number(net.total_rx);
	print_text("\ntotal_mgmt=");
	print_number(net.total_mgmt);
	print_text("\ntotal_data=");
	print_number(net.total_data);
	print_text("\ntotal_udp=");
	print_number(net.total_udp);
	print_text("\ntotal_arp=");
	print_number(net.total_arp);
	print_text("\ntotal_icmp=");
	print_number(net.total_icmp);
	print_text("\ntotal_tcp=");
	print_number(net.total_tcp);
	
}


#else // USE_WIFI


#if defined(USE_USB) && !defined(COPTER_MODE)

void net_start_receive()
{
	usb_start_receive();
	net.current_function = handle_beacon;
}


void net_start_transmit(unsigned char *data, 
	int size, 
	void (*next_function)())
{
	usb_start_transmit(data, size);
	net.current_function = net_wait_tx;
	net.next_function = next_function;
}



// Wait for a transmit to finish, then receive
void net_wait_tx()
{
	if(usb.have_response)
	{
		net.current_function = net.next_function;
	}
}

#endif // defined(USE_USB) && !defined(COPTER_MODE)

#endif  // !USE_WIFI





void init_net()
{


	bzero(&net, sizeof(net));
	
#ifndef USE_WIFI
	
	net.current_function = handle_beacon;
#ifdef PIC_USE_MRF49XA
	net.beacon_offset = COPTER_SERIAL_IN_SIZE;
#endif

	
	
#else // !USE_WIFI

	net.current_function = init_net2;
	net.a_record = kmalloc(strlen(cname_text) + strlen(domain_text) + 3, 0);
	net.ptr_record = kmalloc(4 * 4 + 2, 0);

	memcpy(net.self_ip, this_ip, 4);
	memcpy(net.ground_mac, ground_mac, ETH_ALEN);




	print_text("SSID=");
	print_text(copter.ssid);
	print_text(" Channel=");
	print_number(copter.channel);
	print_lf();



// Copy cname
	int offset = 0;
// strlen screws up analog if init_net comes before init_analog
	int len = strlen(cname_text);
	net.a_record[offset++] = len;
	memcpy(net.a_record + offset, cname_text, len);
	offset += len;

// Copy domain
	const char *ptr = domain_text;
	int count = 0;
	int count_offset = offset;
	offset++;
	while(*ptr != 0)
	{
		while(*ptr != 0 && *ptr != '.')
		{
// copy a character
			net.a_record[offset++] = *ptr++;
			count++;
		}

// Skip period
		if(*ptr == '.') ptr++;
// Store previous count
		net.a_record[count_offset] = count;
		count = 0;
// Record position of next count & skip it
		count_offset = offset;
		offset++;
	}
	net.a_record[count_offset] = 0;
	net.a_record_len = offset;

/*
 * 	TRACE
 * 	print_buffer(net.a_record, net.a_record_len);
 * 	print_text(net.a_record);
 */

// Copy ptr
	offset = 0;
	int i;
	for(i = 0; i < 4; i++)
	{
		len = sprint_number(net.ptr_record + offset + 1, 
			net.self_ip[3 - i],
			4);
		net.ptr_record[offset] = len;
		offset += 1 + len;
	}
	net.ptr_record[offset++] = 0;
	net.ptr_record_len = offset;


// left eye fifo
#ifdef GROUND_MODE
	net.left_fifo = kmalloc(LEFT_FIFO_SIZE, 0);
#endif

/*
 * 	TRACE
 * 	print_buffer(net.ptr_record, net.ptr_record_len);
 * 	print_text(net.ptr_record);
 */

	counter = 0;
#endif // USE_WIFI



}








